/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acx.c,v 1.148 2004/02/26 09:31:27 darron Exp $
 */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxreg.h"
#include "if_acxcmd.h"

/* functions */
#include "if_acx_proto.h"
#include "if_acxreg_proto.h"
#include "if_acxdev_proto.h"
#include "if_acxcmd_proto.h"
#include "if_acxifnet_proto.h"
#include "if_acxtemplate_proto.h"
#include "if_acxwep_proto.h"
#include "if_acxpacket_proto.h"
#include "if_acxsysctl_proto.h"

MODULE_DEPEND(acx, ether, 1, 1, 1);
MODULE_DEPEND(acx, wlan, 1, 1, 1);

/*
 * attach device
 */
int
acx_attach(struct acx_softc *sc)
{
    if( sc->hardware_type == ACX_DEV_111 )
    {
        device_printf(sc->dev, "ACX111 unsupported\n");
        return -1;
    }

    /* attach dma regions and alloc mbufs */
    if( acx_packet_attach(sc) != 0 )
        return -1;

    /* attach device */
    if( acx_dev_attach(sc) != 0 )
        return -1;
    
    /* initialise thread lock XXX UNUSED */
    mtx_init(&sc->lock, device_get_nameunit(sc->dev), "acx lock", MTX_DEF);

    /* initialise command queueing */
    acx_cmd_init(sc);
    
    /* activate device */
    acx_dev_init(sc);

    /* init wep */
    acx_wep_init(sc);

    /* init template */
    acx_template_init(sc);

    /* init packet */
    acx_packet_init(sc);

    /* initialize generic network interface */
    acx_ifnet_attach(sc);

    /* display banner(s) */
    device_printf(sc->dev, "802.11 address: %s\n",
        ether_sprintf( sc->hardware_macaddr ));

    device_printf(sc->dev, "Eeprom Rev %d Domain %s, Firmware %s\n",
        sc->eeprom_version, sc->eeprom_domainstr, sc->firmware_version);
    
    device_printf(sc->dev, "Radio Type 0x%02x, Antenna 0x%02x, CCA Mode 0x%02x, ED Threshold 0x%02x\n",
        sc->hardware_version>>8, sc->current_antenna, sc->cca_mode, sc->ed_threshold);

    device_printf(sc->dev, "(c) http://wlan.kewl.org/ 2003\n");

    return 0;
}

/*
 * detach device
 */
void
acx_detach(struct acx_softc *sc)
{
    /* detach device */
    acx_dev_end(sc);

    /* detach generic network interface */
    acx_ifnet_detach(sc);

    /* deactivate command queue */
    acx_cmd_end(sc);

    /* deactivate thread lock */
    mtx_destroy(&sc->lock);

    /* detach dma regions */
    acx_packet_detach(sc);
}

/*
 * interrupt service routine
 *
 * notes:
 *  do as little as possible here, spawn handlers as required.
 */
void
acx_interrupt(void *scarg)
{
    struct acx_softc *sc= (struct acx_softc *)scarg;
    u_int16_t irqstatus;

    if(sc->irq_enabled == 0)
        return; /* not our interrupt */
    if((irqstatus= READW(&sc->mem1, ACX100_IRQ_STATUS)) == 0 )
        return; /* not our interrupt */
    
    /*
     * interesting irqs
     *
     * notes:
     *  each particular handler shall ack the irq itself, hopefully
     *  to prevent the chip initiating a new irq before we are ready.
     */
    if( irqstatus & 0x0008 )    /* RX */
        taskqueue_enqueue_fast(taskqueue_fast, &sc->packet_task_rx);

    if( irqstatus & 0x0002 )    /* TX */
        taskqueue_enqueue_fast(taskqueue_fast, &sc->packet_task_tx);

    if( irqstatus & 0x0200 )    /* COMMAND COMPLETE */
        taskqueue_enqueue_fast(taskqueue_fast, &sc->cmd_task_irq);

    /*
     * uninteresting or unknown irqs
     *
     * XXX
     * the info irq could be filling in information regarding the last
     * tx packet sent? or packets sent? this infomation would be
     * useful to determine how many tx retries were performed? assuming
     * it give that information. we should inspect the info mailbox
     * scratchpad area to see what we find.
     */
    if( irqstatus & 0x0001 )    /* UNKNOWN */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0001);
    if( irqstatus & 0x0004 )    /* UNKNOWN */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0004);
    if( irqstatus & 0x0010 )    /* DTIM */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0010);
    if( irqstatus & 0x0020 )    /* BEACON */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0020);
    if( irqstatus & 0x0040 )    /* TIMER */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0040);
    if( irqstatus & 0x0080 )    /* KEY NOT FOUND */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0080);
    if( irqstatus & 0x0100 )    /* IV ICV FAILED */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0100);
    if( irqstatus & 0x0400 )    /* INFO */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0400);
    if( irqstatus & 0x0800 )    /* OVERFLOW */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0800);
    if( irqstatus & 0x1000 )    /* ERROR */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x1000);
    if( irqstatus & 0x2000 )    /* SCAN COMPLETE */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x2000);
    if( irqstatus & 0x4000 )    /* BOOT COMPLETE */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x4000);
    if( irqstatus & 0x8000 )    /* UNKNOWN */
        WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x8000);

    return;
}
