/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxcmd.c,v 1.199 2004/02/26 09:31:27 darron Exp $
 */

/****************************************************************************************************************
 * functions to interact with the firmware via the FW register set in order to issue firmware commands          *
 ****************************************************************************************************************/

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxcmd.h"
#include "if_acxreg.h"

/* functions */
#include "if_acxcmd_proto.h"
#include "if_acxreg_proto.h"
#include "if_acxdev_proto.h"

const char
    err0[]= "Idle Timeout",
    err1[]= "OK",
    err2[]= "Unknown Command",
    err3[]= "Invalid Information Element",
    err4[]= "Unknown Error",
    err5[]= "Channel Invalid in Current Regulatory Domain",
    err6[]= "Unknown Error",
    err7[]= "Read Only",
    err8[]= "Command Rejected",
    err9[]= "Already Asleep",
    err10[]="Tx in Progress",
    err11[]="Already Awake",
    err12[]="Write Only",
    err13[]="Rx in Progress",
    err14[]="Invalid Parameter",
    err15[]="Scan in progress",
    err16[]="Unknown Error";

const char *acx_cmd_errors[]= {
    err0,err1,err2,err3,err4,err5,err6,err7,err8,err9,err10,err11,err12,err13,err14,err15,err16
};

/*
 * command queue handler overview
 *
 * with the acx100 firmware api, a command is issued and then
 * an interrupt is generated to signify command completion.
 *
 * a command is entered into the queue and if the queue
 * was previously empty, then this command is sent to the acx100.
 * 
 * if the queue was not empty, then this command is issued later.
 *
 * when the acx100 interrupt occurs to indicate a command completion
 * the topmost queue entry is inspected, and it's callback routine
 * executed or it's thread is woken up. in combination these methods
 * allow both blocking and non blocking command execution.
 *
 * if another entry moves to the top when a command is removed this is
 * then sent to the acx100 for processing to bump the queue along.
 *
 * critical aspects in this process are queue locking. it is important
 * that the queue is locked whilst inserting and deleting entries.
 *
 * during the sysinit stage of freebsd, we are unable to benefit
 * from certain functionality, consequently there exists a third
 * overriding and transparent method of command execution.
 *
 * this scheme is busy block waiting, and will only be used in the init
 * stage prior to IRQs becoming active, calls using callback (non
 * blocking) will not be available.
 */

/*
 * perform init for acx_cmd
 */
void
acx_cmd_init(struct acx_softc *sc)
{
    /* determine mailbox offsets in memory region 2 for firmware transactions
     *
     * command mailbox has the following structure (to be determined)
     *
     * u_int16_t unknown
     * u_int16_t unknown
     * union *   paramaters/results
     *
     * notes:
     *  we ignore the unknown items
     */
    sc->cmdoffset = 4 + READW(&sc->mem1, ACX100_CMD_MAILBOX_OFFS);
    sc->infoffset = 4 + READW(&sc->mem1, ACX100_INFO_MAILBOX_OFFS);
    
    /* initialise command queue */
    STAILQ_INIT(&sc->cmd_head);

    /* initialise command queue exec handler (swi) */
    TASK_INIT( &sc->cmd_task_exec,
        /* priority */ 0,
        /* function */ (task_fn_t *)acx_cmd_task_exec,
        /* context  */ sc);

    /* initialise command queue irq handler (fast) */
    TASK_INIT( &sc->cmd_task_irq,
        /* priority */ 0,
        /* function */ (task_fn_t *)acx_cmd_task_irq,
        /* context  */ sc);

    /* initialise command lock */
    mtx_init(&sc->cmd_lock, device_get_nameunit(sc->dev), "cmd lock", MTX_DEF);
}

void
acx_cmd_end(struct acx_softc *sc)
{
    mtx_destroy(&sc->cmd_lock);
}

/*
 * initialise card for access
 */
inline void
acx_cmd_access(struct acx_softc *sc)
{
    WRITEW(&sc->mem1, ACX100_FW_4, 0x0);
    WRITEW(&sc->mem1, ACX100_FW_5, 0x0);
    WRITEW(&sc->mem1, ACX100_FW_2, 0x0);
    WRITEW(&sc->mem1, ACX100_FW_3, 0x1);
}

/*
 * initialise card for firmware upload
 */
inline void
acx_cmd_upload(struct acx_softc *sc, u_int32_t offset)
{
    acx_cmd_access(sc);
    WRITEW(&sc->mem1, ACX100_FW_0, offset & 0xffff);
    WRITEW(&sc->mem1, ACX100_FW_1, offset >> 16);
}

/*
 * initialise card for get or set info
 */
inline void
acx_cmd_info(struct acx_softc *sc)
{
    acx_cmd_access(sc);
    WRITEW(&sc->mem1, ACX100_FW_0, READW(&sc->mem1, ACX100_CMD_MAILBOX_OFFS));
    WRITEW(&sc->mem1, ACX100_FW_1, 0x0);
}

/*
 * get the command type and status
 */
void
acx_cmd_get_info(struct acx_softc *sc, u_int16_t *cmd_type, u_int16_t *cmd_status, const char **cmd_status_string)
{
    acx_cmd_info(sc);
    *cmd_type  = READW(&sc->mem1, ACX100_DATA_LO);
    *cmd_status= READW(&sc->mem1, ACX100_DATA_HI);

    if( *cmd_status < 16 )
        *cmd_status_string= acx_cmd_errors[*cmd_status];
    else
        *cmd_status_string= acx_cmd_errors[16];
}

/*
 * set the command type
 */
void
acx_cmd_set_type(struct acx_softc *sc, u_int16_t cmd_type)
{
    acx_cmd_info(sc);
    WRITEW(&sc->mem1, ACX100_DATA_LO, cmd_type);
    WRITEW(&sc->mem1, ACX100_DATA_HI, 0x0);
}

/*
 * set the command status
 */
void
acx_cmd_set_status(struct acx_softc *sc, u_int16_t cmd_status)
{
    acx_cmd_info(sc);
    WRITEW(&sc->mem1, ACX100_DATA_LO, 0x0);
    WRITEW(&sc->mem1, ACX100_DATA_HI, cmd_status);
}

/*
 * set the command param
 */
inline void
acx_cmd_set_param(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    bus_space_write_region_1( sc->mem2.bustag, sc->mem2.bushandle,
        sc->cmdoffset, (u_int8_t *) &cmd->param, cmd->param_length );
}

/*
 * get the command result
 */
inline void
acx_cmd_get_result(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    bus_space_read_region_1( sc->mem2.bustag, sc->mem2.bushandle,
        sc->cmdoffset, (u_int8_t *) &cmd->param, cmd->param_length );
}

/*
 * run a command
 */
void
acx_cmd_run(struct acx_softc *sc)
{
    WRITEW(&sc->mem1, ACX100_REG_7C, READW(&sc->mem1, ACX100_REG_7C) | 0x01);
}

/*
 * upload then verify firmware
 *
 * notes:
 *  checksum and size are removed from the firmware image (both verified externally)
 *  firmware array is big endian 32 bit words
 */
int
acx_cmd_upload_firmware(struct acx_softc *sc, struct acx_dev_firmware *fw, u_int32_t offset)
{
    int i;

    /*
     * store firmware image in internal ram
     */
    acx_cmd_upload(sc, offset);
    for(i=0; i<fw->codesize; i++)
    {
        WRITEW(&sc->mem1, ACX100_DATA_LO, fw->code[i] & 0xffff);
        WRITEW(&sc->mem1, ACX100_DATA_HI, fw->code[i] >> 16);
        //DELAY(10);
    }

    /*
     * verify firmware image in internal ram
     *
     * XXX if it doesn't work, expect big problems
     */
    /*
    acx_cmd_upload(sc, offset);
    for(i=0; i<fw->codesize; i++ )
    {
        if( (READW(&sc->mem1, ACX100_DATA_LO) | READW(&sc->mem1, ACX100_DATA_HI) << 16) != fw->code[i] )
            return ACX_CMD_ERR_UPLOADFAILED;
        //DELAY(10);
    }
    */
    return 0;
}

/*
 * setup statistical interface to vmstat -m
 *
 * shared across all instances
 */
MALLOC_DECLARE(ACX_CMD);
MALLOC_DEFINE(ACX_CMD, "ACX CMD", "ACX CMD (an instance of acx_cmd_entry)");

/*
 * create a acx_cmd_entry
 */
struct acx_cmd_entry *
acx_cmd_create(struct acx_softc *sc, u_int16_t command_type, void *callback_func, u_int16_t param_length)
{
    struct acx_cmd_entry *cmd;

    MALLOC(cmd, struct acx_cmd_entry *, sizeof(struct acx_cmd_entry), ACX_CMD, M_WAITOK | M_ZERO);

    if( cmd==NULL )
        panic("acx%d: acx_cmd_create: malloc failed\n", sc->unit);

    cmd->callback_func= callback_func;
    cmd->command_type= command_type;
    cmd->param_length= param_length;

    return cmd;
}
    
/*
 * start a command 
 *
 * notes:
 *  should be called inside command lock
 */
void
acx_cmd_start(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    /* determine if a command is still in progress */
    if( sc->cmd!=NULL )
        panic("acx%d: acx_cmd_start: command in progress", sc->unit);

    /* save command block */
    sc->cmd= cmd;
    
    /* set up parameter mailbox */
    if(cmd->param_length > 0)
        acx_cmd_set_param(sc, cmd);

    /* set the command type */
    acx_cmd_set_type(sc, cmd->command_type);

    /* run the command in the firmware */
    acx_cmd_run(sc);
}

/*
 * process command start
 */
void
acx_cmd_task_exec(struct acx_softc *sc, int count)
{
    struct acx_cmd_entry *cmd;

    mtx_lock(&sc->cmd_lock);
    
    /* determine pending command */
    if( (cmd= STAILQ_FIRST(&sc->cmd_head)) == NULL )
        panic("acx%d: acx_cmd_task_exec: command queue empty", sc->unit);
    
    /* start the command */
    acx_cmd_start(sc, cmd);
    
    mtx_unlock(&sc->cmd_lock);
}

/*
 * complete a command
 *
 * notes:
 *  should be called inside command lock
 */
void
acx_cmd_complete(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    u_int16_t command_type;

    /* inspect running command */
    if(cmd == NULL)
        panic("acx%d: acx_cmd_complete: no command in progress\n", sc->unit);
    
    /* determine status of command */
    acx_cmd_get_info(sc, &command_type, &cmd->command_status, &cmd->command_status_string);
    if( cmd->command_status!=1 )
        panic("acx%d: acx_cmd_task_exec: command failed (1!=%d) [%s]\n", sc->unit,
            cmd->command_status, cmd->command_status_string);
    
    /* save command results */
    acx_cmd_get_result(sc, cmd);

    /* clear status flag */
    acx_cmd_set_status(sc, 0);

    /* clear command block */
    sc->cmd= NULL;
    
    /* acknowledge command completion IRQ */
    WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x200);
}

/*
 * process command completion
 */
void
acx_cmd_task_irq(struct acx_softc *sc, int count)
{
    struct acx_cmd_entry *cmd, *top;
    
    mtx_lock(&sc->cmd_lock);
    
    /* determine current command */
    cmd= sc->cmd;

    /* perform command completion */
    acx_cmd_complete(sc, cmd);
    
    /* remove command from queue */
    STAILQ_REMOVE_HEAD(&sc->cmd_head, acx_cmd_entries);

    /* determing if another command is pending */
    top= STAILQ_FIRST(&sc->cmd_head);

    mtx_unlock(&sc->cmd_lock);
    
    /* wakeup blocking thread or call callback for non blocking */
    if( cmd->callback_func==ACX_CMD_BLOCK )
        wakeup(cmd);
    else
        cmd->callback_func(sc, cmd);

    /* line up another command in the queue */
    if( top!=NULL )
        taskqueue_enqueue(taskqueue_swi, &sc->cmd_task_exec);
}

/*
 * enter a command into the queue or run it now if
 * irqs not available
 */
void
acx_cmd_issue(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    struct acx_cmd_entry *top;
    int error;

    mtx_lock(&sc->cmd_lock);

    if( sc->irq_enabled )   /* irqs available */
    {
        top= STAILQ_FIRST(&sc->cmd_head);

        /* insert command into queue */
        STAILQ_INSERT_TAIL(&sc->cmd_head, cmd, acx_cmd_entries);
        if( top==NULL )
            taskqueue_enqueue(taskqueue_swi, &sc->cmd_task_exec);

        /* block or return */
        if( cmd->callback_func==ACX_CMD_BLOCK )
        {
            /* msleep:
             *
             * 1) release lock
             * 2) sleep
             * 3) reacquire lock
             */
            if( (error= msleep(cmd, &sc->cmd_lock, 0, "msleep", ACX_CMD_BLOCK_TIMEOUT)) != 0)
                panic("acx%d: command timeout\n", sc->unit); /* we need more bug reports, so panic */
        }
    }
    /* irqs unavailble */
    else if( cmd->callback_func==ACX_CMD_BLOCK )
    {
        /* start command */
        acx_cmd_start(sc, cmd);

        /* wait for firmware */
        error=0;
        while( !(READW(&sc->mem1, ACX100_STATUS) & 0x0200) )
        {
            if( error++ == 500 ) /* 5 seconds */
                panic("acx%d: command timeout\n", sc->unit);

            DELAY(10000); /* 10ms */
        }

        /* complete command */
        acx_cmd_complete(sc, cmd);
    }
    else panic("acx%d: non blocking command disallowed in this context\n", sc->unit);

    mtx_unlock(&sc->cmd_lock);
}

/*********************
 * firmware functions
 *********************/

/**************************
 * CONFIGURE - SET COMMANDS
 **************************/

/*
 * set timer
 *
 * notes:
 *  doesn't work on latest firmware, so is not implemented.
 */
void
acx_cmd_set_timer(struct acx_softc *sc)
{
    device_printf(sc->dev, "acx_cmd_set_timer: problem: function unimplemented\n");
}
void
acx_cmd_set_timer_callback(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    device_printf(sc->dev, "acx_cmd_set_timer_callback: problem: function unimplemented\n");
}

/*
 * set memory map
 *
 * notes:
 *  pass in pointer to a memory map structure to be copied out
 */
void
acx_cmd_set_memory_map(struct acx_softc *sc, struct acx_cmd_memory_map_param *memory_map)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_MEMORY_MAP;
    cmd->param.resource.length  = sizeof(struct acx_cmd_memory_map_param);
    memcpy( &cmd->param.resource.param.memory_map, memory_map, sizeof(struct acx_cmd_resource_param));
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set wep options
 *
 * notes:
 *  mode=0 (ACX_WEP_HARDWARE)    wep decrypt in hardware
 *  mode=2 (ACX_WEP_SOFTWARE) no wep decrypt in hardware (passthrough)
 *
 *  XXX this doesn't appear to work correctly.
 */
void
acx_cmd_set_wep_options(struct acx_softc *sc, u_int8_t mode)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_WEP_OPTIONS;
    cmd->param.resource.length  = sizeof(struct acx_cmd_wep_options_param);
    cmd->param.resource.param.wep_options.word0= 0x0e;
    cmd->param.resource.param.wep_options.mode = mode;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set CCA (clear channel assessment) mode
 *
 * notes:
 *  the CCA mode is used to determine when the send data across a channel
 */
void
acx_cmd_set_current_cca_mode(struct acx_softc *sc, u_int8_t mode)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_CURRENT_CCA_MODE;
    cmd->param.resource.length  = sizeof(struct acx_cmd_current_cca_param);
    cmd->param.resource.param.current_cca.mode = mode;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set ED (energy detect) threshold
 *
 * notes:
 *  probably a filter than ignores signals under a specific strength
 */
void
acx_cmd_set_ed_threshold(struct acx_softc *sc, u_int8_t threshold)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_ED_THRESHOLD;
    cmd->param.resource.length  = sizeof(struct acx_cmd_ed_threshold_param);
    cmd->param.resource.param.ed_threshold.level = threshold;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set current regulatory domain
 */
void
acx_cmd_set_current_reg_domain(struct acx_softc *sc, u_int8_t domain)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_CURRENT_REG_DOMAIN;
    cmd->param.resource.length  = sizeof(struct acx_cmd_current_reg_param);
    cmd->param.resource.param.current_reg.domain= domain;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set current antenna selection
 */
void
acx_cmd_set_current_antenna(struct acx_softc *sc, u_int8_t select)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_CURRENT_ANTENNA;
    cmd->param.resource.length  = sizeof(struct acx_cmd_current_antenna_param);
    cmd->param.resource.param.current_antenna.selection= select;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set ssid
 * 
 * notes:
 *  we probably don't need this.
 */
void
acx_cmd_set_ssid(struct acx_softc *sc, char *nwid, int nwidlen)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_SSID;
    cmd->param.resource.length  = nwidlen /* sizeof(struct acx_cmd_ssid_param ) */;
    memcpy( &cmd->param.resource.param.ssid.ssid, nwid, nwidlen /* IEEE80211_NWID_LEN */);
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set rx options
 */
void
acx_cmd_set_rx_config(struct acx_softc *sc, u_int32_t options)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_RXCONFIG;
    cmd->param.resource.length  = sizeof(struct acx_cmd_rx_config_param);
    cmd->param.resource.param.rx_config.options= options;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set firmware queue config
 *
 * notes:
 *  pass in a queue config structure to setup
 */
void
acx_cmd_set_queue_config(struct acx_softc *sc, struct acx_cmd_queue_config_param *queue_config)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_QUEUE_CONFIG;
    cmd->param.resource.length  = sizeof(struct acx_cmd_queue_config_param);
    memcpy( &cmd->param.resource.param.queue_config, queue_config, sizeof(struct acx_cmd_queue_config_param) );
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set memory config options
 *
 * notes:
 *  pass in a memory config options structure to set
 */
void
acx_cmd_set_memory_config_options(struct acx_softc *sc, struct acx_cmd_memory_config_options_param *memory_options)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_MEMORY_CONFIG_OPTIONS;
    cmd->param.resource.length  = sizeof(struct acx_cmd_memory_config_options_param);
    memcpy( &cmd->param.resource.param.memory_config_options, memory_options, sizeof(struct acx_cmd_memory_config_options_param) );
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set block size for dma
 */
void
acx_cmd_set_block_size(struct acx_softc *sc, u_int16_t block_size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_BLOCK_SIZE;
    cmd->param.resource.length  = sizeof(struct acx_cmd_block_size_param);
    cmd->param.resource.param.block_size.size= block_size;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set msdu lifetime
 */
void
acx_cmd_set_max_xmit_msdu_lifetime(struct acx_softc *sc, u_int32_t lifetime)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_MAX_XMIT_MSDU_LIFETIME;
    cmd->param.resource.length  = sizeof(struct acx_cmd_max_xmit_msdu_lifetime_param);
    cmd->param.resource.param.lifetime.lifetime= lifetime;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set short retry
 */
void
acx_cmd_set_short_retry_limit(struct acx_softc *sc, u_int8_t limit)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_SHORT_RETRY_LIMIT;
    cmd->param.resource.length  = sizeof(struct acx_cmd_short_retry_limit_param);
    cmd->param.resource.param.short_limit.limit= limit;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set long retry
 */
void
acx_cmd_set_long_retry_limit(struct acx_softc *sc, u_int8_t limit)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_LONG_RETRY_LIMIT;
    cmd->param.resource.length  = sizeof(struct acx_cmd_long_retry_limit_param);
    cmd->param.resource.param.long_limit.limit= limit;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/* set firmware version */
    
void
acx_cmd_set_wep_tx_key(struct acx_softc *sc, u_int8_t txkey)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_WEP_DEFAULT_KEY_SET;
    cmd->param.resource.length= sizeof(struct acx_cmd_wep_key_default_param);
    cmd->param.resource.param.wep_key_default.index= txkey;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * set wep key
 */
void
acx_cmd_set_wep_key(struct acx_softc *sc, u_int8_t index)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIGURE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_WEP_KEY;
    cmd->param.resource.length= sizeof(struct acx_cmd_wep_key_param);

    cmd->param.resource.param.wep_key.byte0= 1;
    cmd->param.resource.param.wep_key.length=  sc->wepkey[index].len;
    cmd->param.resource.param.wep_key.index= index;
    memcpy( &cmd->param.resource.param.wep_key.key, sc->wepkey[index].key, ACX_WEP_KEYSIZE );
    acx_cmd_issue(sc, cmd);

    FREE(cmd, ACX_CMD);
}

/* set station id */

/*****************************
 * INTERROGATE - GET COMMANDS
 *****************************/

/* get timer */

/*
 * get memory map
 *
 * notes:
 *  pass in pointer to a memory map structure to be copied in
 */
void
acx_cmd_get_memory_map(struct acx_softc *sc, struct acx_cmd_memory_map_param *memory_map)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_MEMORY_MAP;
    cmd->param.resource.length  = sizeof(struct acx_cmd_memory_map_param);
    acx_cmd_issue(sc, cmd);
    memcpy( memory_map, &cmd->param.resource.param.memory_map, sizeof(struct acx_cmd_memory_map_param));
    FREE(cmd, ACX_CMD);
}

/*
 * get wep options
 *
 * notes:
 *  not implemented, wep options are WRITE ONLY, no read access is allowed
 */

/*
 * get CCA (clear channel assessment) mode
 */
void
acx_cmd_get_current_cca_mode(struct acx_softc *sc, u_int8_t *mode)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_CURRENT_CCA_MODE;
    cmd->param.resource.length  = sizeof(struct acx_cmd_current_cca_param);
    acx_cmd_issue(sc, cmd);
    *mode= cmd->param.resource.param.current_cca.mode;
    FREE(cmd, ACX_CMD);
}

/*
 * get ED (energy detect) threshold
 */
void
acx_cmd_get_ed_threshold(struct acx_softc *sc, u_int8_t *threshold)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_ED_THRESHOLD;
    cmd->param.resource.length  = sizeof(struct acx_cmd_ed_threshold_param);
    acx_cmd_issue(sc, cmd);
    *threshold= cmd->param.resource.param.ed_threshold.level;
    FREE(cmd, ACX_CMD);
}

/* get current regulatory domain */

/*
 * get current antenna selection
 */
void
acx_cmd_get_current_antenna(struct acx_softc *sc, u_int8_t *select)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_CURRENT_ANTENNA;
    cmd->param.resource.length  = sizeof(struct acx_cmd_current_antenna_param);
    acx_cmd_issue(sc, cmd);
    *select= cmd->param.resource.param.current_antenna.selection;
    FREE(cmd, ACX_CMD);
}

/* get ssid */

/*
 * get rx options
 */
void
acx_cmd_get_rx_config(struct acx_softc *sc, u_int32_t *options)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_RXCONFIG;
    cmd->param.resource.length  = sizeof(struct acx_cmd_rx_config_param);
    acx_cmd_issue(sc, cmd);
    *options= cmd->param.resource.param.rx_config.options;
    FREE(cmd, ACX_CMD);
}

/* get firmware queue config */
/* get memory config options */
/* get block size for dma */

/*
 * get msdu lifetime
 */
void
acx_cmd_get_max_xmit_msdu_lifetime(struct acx_softc *sc, u_int32_t *lifetime)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_MAX_XMIT_MSDU_LIFETIME;
    cmd->param.resource.length  = sizeof(struct acx_cmd_max_xmit_msdu_lifetime_param);
    acx_cmd_issue(sc, cmd);
    *lifetime= cmd->param.resource.param.lifetime.lifetime;
    FREE(cmd, ACX_CMD);
}

/*
 * get short retry
 */
void
acx_cmd_get_short_retry_limit(struct acx_softc *sc, u_int8_t *limit)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_SHORT_RETRY_LIMIT;
    cmd->param.resource.length  = sizeof(struct acx_cmd_short_retry_limit_param);
    acx_cmd_issue(sc, cmd);
    *limit= cmd->param.resource.param.short_limit.limit;
    FREE(cmd, ACX_CMD);
}

/*
 * get long retry
 */
void
acx_cmd_get_long_retry_limit(struct acx_softc *sc, u_int8_t *limit)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_LONG_RETRY_LIMIT;
    cmd->param.resource.length  = sizeof(struct acx_cmd_long_retry_limit_param);
    acx_cmd_issue(sc, cmd);
    *limit= cmd->param.resource.param.long_limit.limit;
    FREE(cmd, ACX_CMD);
}

/*
 * get firmware version
 */
void
acx_cmd_get_fwrev(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_FWREV;
    cmd->param.resource.length= sizeof(struct acx_cmd_firmware_version_param);
    acx_cmd_issue(sc, cmd);
    strncpy( sc->firmware_version, cmd->param.resource.param.firmware_version.version, sizeof(struct acx_cmd_firmware_version_param));
    FREE(cmd, ACX_CMD);
}

/* get wep keys */

/*
 * get station id
 *
 * notes:
 *  the firmware mac address if low->high
 */
void
acx_cmd_get_station_id(struct acx_softc *sc)
{
    int i=5, j=0;
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INTERROGATE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_resource_param));

    cmd->param.resource.type= ACX100_RID_DOT11_STATION_ID;
    cmd->param.resource.length  = sizeof(struct acx_cmd_station_id_param);
    acx_cmd_issue(sc, cmd);
    while(i>=0)
        sc->hardware_macaddr[j++]= cmd->param.resource.param.macaddr.id[i--];
    FREE(cmd, ACX_CMD);
}

/*****************
 * OTHER COMMANDS
 *****************/

/*
 * sleep
 */
void
acx_cmd_sleep(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_SLEEP, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * wake
 */
void
acx_cmd_wake(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_WAKE, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * initialise memory
 *
 * notes:
 *  firmware initialises descriptors etc internally.
 */
void
acx_cmd_init_memory(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_INIT_MEMORY, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * join network
 *
 * notes:
 *  we probably don't need this.
 *  calling this function affects interrupts in some way.
 */
void
acx_cmd_join_nb(struct acx_softc *sc, u_int8_t macmode)
{
    struct ieee80211com *ic;
    struct ieee80211_node *ni;
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_JOIN, acx_cmd_generic_callback, sizeof(struct acx_cmd_join_param));
    int i;

    /* get info */
    ic= &sc->ic;
    ni= ic->ic_bss;

    /* investigate these */
    cmd->param.join.beacon_interval= 100;
    cmd->param.join.dtim_interval  = 2;
    cmd->param.join.rates_basic    = 15;
    cmd->param.join.rates_supported= 31;
    cmd->param.join.rate_tx= 20;
    cmd->param.join.preamble_type= 0;

    /* set the mac mode */
    cmd->param.join.macmode= macmode;
    
    /* set the bss id */
    for(i=0; i<=5; i++)
        cmd->param.join.bssid[5-i]= ni->ni_bssid[i];

    /* set the ssid network id */
    memcpy(&cmd->param.join.essid, ni->ni_essid, ni->ni_esslen);
    cmd->param.join.essid_len= ni->ni_esslen;

    /* set the channel */
    cmd->param.join.channel=
        ieee80211_mhz2ieee(ic->ic_ibss_chan->ic_freq, ic->ic_ibss_chan->ic_flags);
    
    acx_cmd_issue(sc, cmd);
}

/*
 * start scan
 *
 * notes:
 *  we probably don't need this.
 */
void
acx_cmd_scan(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_SCAN, ACX_CMD_BLOCK, sizeof(struct acx_cmd_scan_param));

    cmd->param.scan.count=              1;
    cmd->param.scan.start_chan=         sc->eeprom_domainlow;
    cmd->param.scan.flags=              0x8000;
    cmd->param.scan.max_rate=           20;     /* 2 Mbps */
    cmd->param.scan.options=            1;
    cmd->param.scan.chan_duration=      500;
    cmd->param.scan.max_probe_delay=    200;

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * stop scan
 *
 * notes:
 *  we probably don't need this.
 */
void
acx_cmd_stop_scan(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_STOP_SCAN, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * enable receive (non blocking)
 *
 * notes:
 *  channel is the ieee channel number to receive on.
 *  calling rxconfig before enabling rx seems to cause problems in the firmware.
 */
void
acx_cmd_enable_rx_nb(struct acx_softc *sc, u_int8_t channel)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_ENABLE_RX, acx_cmd_generic_callback, sizeof(struct acx_cmd_enable_rx_param));

    cmd->param.enable_rx.channel = channel;
    acx_cmd_issue(sc, cmd);
}

/*
 * disable receive
 */
void
acx_cmd_disable_rx(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_DISABLE_RX, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * enable transmit (non blocking)
 *
 * notes:
 *  channel is the ieee channel number to transmit on.
 */
void
acx_cmd_enable_tx_nb(struct acx_softc *sc, u_int8_t channel)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_ENABLE_TX, acx_cmd_generic_callback, sizeof(struct acx_cmd_enable_tx_param));

    cmd->param.enable_tx.channel = channel;
    acx_cmd_issue(sc, cmd);
}

/*
 * disable transmit
 */
void
acx_cmd_disable_tx(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_DISABLE_TX, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * radio init
 *
 * notes:
 *  initialise the radio after it's firmware has been uploaded.
 */
void
acx_cmd_radioinit(struct acx_softc *sc, u_int32_t radio_offset, u_int32_t radio_size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_RADIOINIT, ACX_CMD_BLOCK, sizeof(struct acx_cmd_radioinit_param));

    cmd->param.radioinit.offset= radio_offset;
    cmd->param.radioinit.length= radio_size;
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * config beacon
 *
 * notes:
 *  we probably don't need this.
 *  call with empty packet to initialise
 */
void
acx_cmd_config_beacon(struct acx_softc *sc, u_int8_t *packet, int size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIG_BEACON, ACX_CMD_BLOCK, sizeof(struct acx_cmd_config_beacon_param));

    cmd->param.config_beacon.size= size;
    memcpy( &cmd->param.config_beacon.packet, packet, size);
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * config tim
 *
 * notes:
 *  we probably don't need this.
 *  call with empty template to initialise
 */
void
acx_cmd_config_tim(struct acx_softc *sc, u_int8_t *packet, int size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIG_TIM, ACX_CMD_BLOCK, sizeof(struct acx_cmd_config_tim_param));

    cmd->param.config_tim.size= size;
    memcpy( &cmd->param.config_tim.packet, packet, size);
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * config probe response
 *
 * notes:
 *  we probably don't need this.
 *  call with empty packet to initialise
 */
void
acx_cmd_config_probe_response(struct acx_softc *sc, u_int8_t *packet, int size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIG_PROBE_RESPONSE, ACX_CMD_BLOCK, sizeof(struct acx_cmd_config_probe_response_param));

    cmd->param.config_probe_response.size= size;
    memcpy( &cmd->param.config_probe_response.packet, packet, size);
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * config probe request
 *
 * notes:
 *  we probably don't need this.
 *  initialisation is not needed
 */
void
acx_cmd_config_probe_request(struct acx_softc *sc, u_int8_t *packet, int size)
{
    struct acx_cmd_entry *cmd=
        acx_cmd_create(sc, ACX100_CMD_CONFIG_PROBE_REQUEST, ACX_CMD_BLOCK, sizeof(struct acx_cmd_config_probe_request_param));

    cmd->param.config_probe_request.size= size;
    memcpy( &cmd->param.config_probe_request.packet, packet, size);
    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * reset request
 *
 * notes:
 *  no idea what this does, hoping we don't need to re-init.
 *  this may clear the descriptors.
 */
void
acx_cmd_reset(struct acx_softc *sc)
{
    struct acx_cmd_entry *cmd =
        acx_cmd_create(sc, ACX100_CMD_RESET, ACX_CMD_BLOCK, 0);

    acx_cmd_issue(sc, cmd);
    FREE(cmd, ACX_CMD);
}

/*
 * generic command callback that just cleans up
 */
void
acx_cmd_generic_callback(struct acx_softc *sc, struct acx_cmd_entry *cmd)
{
    FREE(cmd, ACX_CMD);
}
