/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxdev.c,v 1.59 2004/02/15 02:30:22 darron Exp $
 */

/*
 * functions to initialise the hardware and software on the device
 */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxreg.h"
#include "if_acxeeprom.h"
#include "if_acxcmd.h"
#include "if_acxdev.h"

/* functions */
#include "if_acxdev_proto.h"
#include "if_acxreg_proto.h"
#include "if_acxeeprom_proto.h"
#include "if_acxcmd_proto.h"

/*
 * firmware
 */
#include "firmware/wlangen.h" /* ACX100 */
struct acx_dev_firmware acx100_firmware=    { sizeof(wlangen_code)/4, wlangen_code };
#include "firmware/radio0d.h" /* MAXIM */
struct acx_dev_firmware acx100_radio0d=     { sizeof(radio0d_code)/4, radio0d_code };
#include "firmware/radio11.h" /* RFMD */
struct acx_dev_firmware acx100_radio11=     { sizeof(radio11_code)/4, radio11_code };
#include "firmware/radio15.h" /* RADIO15 */
struct acx_dev_firmware acx100_radio15=     { sizeof(radio15_code)/4, radio15_code };

/*
 * #include "firmware/tiacx111.h"
 * struct acx_dev_firmware acx111_firmware=    { sizeof(tiacx111_code)/4, tiacx111_code };
 */

/*
 * attach device hardware and software
 *
 * notes:
 *  failure here results in the device not being attached.
 */
int
acx_dev_attach(struct acx_softc *sc)
{
    u_int16_t temp=0;

    /*
     * set bit0 of RESET_2
     */
    WRITEW(&sc->mem1, ACX100_RESET_2, READW(&sc->mem1, ACX100_RESET_2) | 0x1 );

    /*
     * toggle bit0 of REG0
     */
    temp= READW(&sc->mem1, ACX100_REG0) | 0x1;
    WRITEW(&sc->mem1, ACX100_REG0, temp);
    DELAY(10000); /* 10ms */
    WRITEW(&sc->mem1, ACX100_REG0, temp & 0xfffe);

    /*
     * set bit0 of RESET_0
     */
    WRITEW(&sc->mem1, ACX100_RESET_0, READW(&sc->mem1, ACX100_RESET_0) | 0x1 );
    DELAY(10000); /* 10ms */

    temp= READW(&sc->mem1, ACX100_RESET_2) & 1;
    if( temp == 0 )
        return -1;

    if( READW(&sc->mem1, ACX100_RESET_1) & 2)
        return -1;
    
    /*
     * upload firmware
     */
    if( acx_cmd_upload_firmware(sc, &acx100_firmware, 0) != 0)
        return -1;

    WRITEW(&sc->mem1, ACX100_RESET_2, temp & 0xfffe);

    /* wait for firmware */
    temp=0;
    while( !(READW(&sc->mem1, ACX100_STATUS) & 0x4000) )
    {
        if( temp++ == 500 ) /* 5 seconds */
            return -1;
        DELAY(10000); /* 10ms */
    }

    /* acknowledge boot irq */
    WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x4000);

    /* determine radio version */
    sc->hardware_version= READW(&sc->mem1, 0x2ac);

    /* determine eeprom version/domain */
    acx_eeprom_init(sc);
    
    return 0;
}

/*
 * power LED control
 */
void
acx_dev_power_led_on(struct acx_softc *sc)
{
    WRITEW(&sc->mem1, 0x290, READW(&sc->mem1, 0x290) & ~0x0800);
}
void
acx_dev_power_led_off(struct acx_softc *sc)
{
    WRITEW(&sc->mem1, 0x290, READW(&sc->mem1, 0x290) | 0x0800);
}

/*
 * irq control
 */
void
acx_dev_irq_on(struct acx_softc *sc)
{
    WRITEW(&sc->mem1, ACX100_IRQ_MASK, ACX_DEV_IRQON);
    WRITEW(&sc->mem1, ACX100_IRQ_34, 0x8000);
    sc->irq_enabled=1;
}
void
acx_dev_irq_off(struct acx_softc *sc)
{
    sc->irq_enabled=0;
    WRITEW(&sc->mem1, ACX100_IRQ_MASK, ACX_DEV_IRQOFF);
    WRITEW(&sc->mem1, ACX100_IRQ_34, 0x7FFF);
}

/*
 * initialise the device
 */
void
acx_dev_init(struct acx_softc *sc)
{
    u_int32_t new_offset, radio_offset, radio_size=0;
    u_int8_t radio_type= sc->hardware_version>>8;
    struct acx_cmd_memory_map_param *memory_map;

    /* activate power LED */
    acx_dev_power_led_on(sc);

    /*
     * upload radio
     */
    MALLOC(memory_map, struct acx_cmd_memory_map_param *,
        sizeof(struct acx_cmd_memory_map_param), M_TEMP, M_WAITOK | M_ZERO);

    acx_cmd_get_memory_map(sc, memory_map);
    radio_offset= memory_map->CodeEnd;

    acx_cmd_sleep(sc);

    if( radio_type == 0x0d )
    {
        acx_cmd_upload_firmware(sc, &acx100_radio0d, radio_offset);
        radio_size= (u_int32_t) acx100_radio0d.codesize * 4;
    }
    else if( radio_type == 0x11 )
    {
        acx_cmd_upload_firmware(sc, &acx100_radio11, radio_offset);
        radio_size= (u_int32_t) acx100_radio11.codesize * 4;
    }
    else if( radio_type == 0x15 )
    {
        acx_cmd_upload_firmware(sc, &acx100_radio15, radio_offset);
        radio_size= (u_int32_t) acx100_radio15.codesize * 4;
    }

    acx_cmd_wake(sc);
    acx_cmd_radioinit(sc, radio_offset, radio_size);

    acx_cmd_get_memory_map(sc, memory_map);
    new_offset= memory_map->CodeEnd;

    if( new_offset != ( radio_offset + radio_size ) )
        device_printf(sc->dev, "acx_dev_activate: problem: new_offset incorrect\n");

    FREE(memory_map, M_TEMP);

    /*
     * determine MAC address
     */
    acx_cmd_get_station_id(sc);

    /*
     * determine firmware version
     */
    acx_cmd_get_fwrev(sc);

    /*
     * ensure correct regional domain
     */
    acx_cmd_set_current_reg_domain(sc, sc->eeprom_domain);

    /*
     * set as card defaults
     *
     * XXX defaults on the WL-1000 and USR 2216
     */
    /*
    acx_cmd_set_current_cca_mode(sc, 0x0d);
    acx_cmd_set_ed_threshold(sc, 0x70);
    acx_cmd_set_current_antenna(sc, 0x8F);
    */
    acx_cmd_get_current_cca_mode(sc, &sc->cca_mode);
    acx_cmd_get_ed_threshold(sc, &sc->ed_threshold);
    acx_cmd_get_current_antenna(sc, &sc->current_antenna);

    /*
     * set to experimental values
     */
    acx_cmd_set_max_xmit_msdu_lifetime(sc, 0x2000);
    acx_cmd_set_short_retry_limit(sc, 16);
    acx_cmd_set_long_retry_limit(sc, 16);

    /* does this make a difference? */
    acx_dev_set_power_level(sc, 63);

    /* null receiver config */
    acx_cmd_set_rx_config(sc, 0);
}

/*
 * deactivate the device
 */
void
acx_dev_end(struct acx_softc *sc)
{
    /*
     * deactivate power LED
     */
    acx_dev_power_led_off(sc);

    /*
     * deactivate interrupts
     */
    acx_dev_irq_off(sc);
}

/*
 * set the hardware tx power level (0..63)
 */
void
acx_dev_set_power_level(struct acx_softc *sc, u_int8_t power_level)
{
    u_int8_t radio_type= sc->hardware_version>>8;

    switch(radio_type)
    {
        case 0x0d: /* maxim */
                power_level= 0x3F - ( power_level & 0x3F);
                break;

        case 0x11: /* rfmd */
                power_level= power_level & 0x3F;
                break;

        default:   /* unknown */
                return;
    }

    WRITEW(&sc->mem1, 0x268, 0x11);
    WRITEW(&sc->mem1, 0x26a, 0);
    WRITEW(&sc->mem1, 0x26c, power_level);
    WRITEW(&sc->mem1, 0x26e, 0);
    WRITEW(&sc->mem1, 0x270, 1);
    WRITEW(&sc->mem1, 0x272, 0);
}
