/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxifnet.c,v 1.145 2004/02/26 10:48:01 darron Exp $
 */

/*
 * generic network interface
 *
 * XXX locking should be done in here
 * XXX determine really what to do on ifmedia change
 *     (why doesn't this always register correctly in ifconfig)
 *     (what are the parameters really representing)
 */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxcmd.h"
#include "if_acxioctl.h"
#include "if_acxpacket.h"
#include "if_acxifnet.h"

/* functions */
#include "if_acxifnet_proto.h"
#include "if_acxcmd_proto.h"
#include "if_acxioctl_proto.h"
#include "if_acxpacket_proto.h"
#include "if_acxdev_proto.h"
#include "if_acxutil_proto.h"

/* global sysctl vars */
int acx_ifnet_scan_period= ACX_IFNET_SCANPERIOD;

/*
 * attach a generic network interface to the system
 */
void
acx_ifnet_attach(struct acx_softc *sc)
{
    struct ieee80211com *ic;
    struct ifnet *ifp;

    /* extract some useful vars */
    ic = &sc->ic;
    ifp = &ic->ic_if;

    /* lock */
    mtx_init(&sc->ifnet_lock, device_get_nameunit(sc->dev), "ifnet lock", MTX_DEF);

    /* initialise scan callout without a lock */
    callout_init(&sc->ifnet_callout_scan, 0 /*CALLOUT_MPSAFE*/);

    /* setup generic interface */
    if_initname(ifp, device_get_name(sc->dev), device_get_unit(sc->dev));
    ifp->if_softc = sc;

    /*
     * XXX
     *
     * i have no idea how to deal with multicast in the driver firmware
     *
     * we could set the read config to promiscuous and then write a software
     * filter feeding ieee input.
     *
     * info: http://www.hep.ucl.ac.uk/~ytl/multi-cast/addresstranslation_01.html
     */
    ifp->if_flags = IFF_SIMPLEX | IFF_BROADCAST | IFF_MULTICAST;

    /* if_start - initiate output */
    ifp->if_ioctl = acx_ifnet_ioctl;

    ifp->if_watchdog = acx_packet_watchdog;

    /* if_init - should (re-)init card (set channel too) */
    ifp->if_init = acx_ifnet_init;

    /* send packet(s) */
    ifp->if_start = acx_packet_tx;

    /* setup ieee80211 interface */
    acx_ifnet_setchannels(ic, sc->eeprom_domainlow, sc->eeprom_domainhigh);
    acx_ifnet_setrates(ic);

    ifp->if_snd.ifq_maxlen = IFQ_MAXLEN; /* 50 */

    ic->ic_phytype= IEEE80211_T_DS;
    ic->ic_opmode = IEEE80211_M_STA;
    ic->ic_state  = IEEE80211_S_INIT;

    /* allow adhoc even though it doesn't work */

    ic->ic_caps   = IEEE80211_C_WEP |         /* wep supported */
                  IEEE80211_C_IBSS |        /* ibss/adhoc modes */
//                  IEEE80211_C_HOSTAP |      /* hostap mode */
//                  IEEE80211_C_MONITOR |     /* monitor mode */
                    IEEE80211_C_SHPREAMBLE  /* short preamble supported */
               /*   IEEE80211_C_RCVMGT */ ;     /* recv management frames */

    IEEE80211_ADDR_COPY(ic->ic_myaddr, sc->hardware_macaddr);

    ieee80211_ifattach(ifp);
    ieee80211_media_init(ifp, ieee80211_media_change, ieee80211_media_status);

    sc->ic_newstate= ic->ic_newstate;
    ic->ic_newstate= acx_ifnet_newstate;

}

/*
 * detach interface
 */
void
acx_ifnet_detach(struct acx_softc *sc)
{
    struct ifnet *ifp;

    /*
     * cancel callout
     *
     * notes:
     *  callout_stop returns 0 if the callout is finished or in operation
     *  so we delay a while to hope we don't exit too early
     */
    callout_stop(&sc->ifnet_callout_scan);
    tsleep(sc, 0, "tsleep", hz);
    
    ifp = &sc->ic.ic_if;
    ieee80211_ifdetach(ifp);

    mtx_destroy(&sc->ifnet_lock);
}

/*
 * start scanning for an AP
 */
void
acx_ifnet_start_scan(struct ifnet *ifp)
{
    struct acx_softc *sc= ifp->if_softc;
    struct ieee80211com *ic= &sc->ic;
    u_int8_t channel;

    /* start scanning */
    ieee80211_begin_scan(ifp);

    /* set the initial channel (non block) */
    channel= ieee80211_chan2ieee(ic, ic->ic_bss->ni_chan);
    acx_cmd_enable_rx_nb(sc, channel);
    acx_cmd_enable_tx_nb(sc, channel);
    
    /* move to next channel later */
    callout_reset(&sc->ifnet_callout_scan, acx_ifnet_scan_period, acx_ifnet_callout_next_scan, (void *)ifp);
}

/*
 * move to next frequency for scan
 *
 * notes:
 *  this routine can preempt anything, and must be treated
 *  the same as an IRQ.
 */
void
acx_ifnet_callout_next_scan(void *arg)
{
    struct ifnet *ifp= (struct ifnet *) arg;
    struct acx_softc *sc= ifp->if_softc;
    struct ieee80211com *ic= &sc->ic;
    u_int8_t channel;

    if( ic->ic_state == IEEE80211_S_SCAN )
    {
        /* keep scanning */
        ieee80211_next_scan(ifp);

        /* set the channel we are scanning on (non blocking) */
        channel= ieee80211_chan2ieee(ic, ic->ic_bss->ni_chan);
        acx_cmd_enable_rx_nb(sc, channel);
        acx_cmd_enable_tx_nb(sc, channel);
        
        /* move to next channel later */
        callout_reset(&sc->ifnet_callout_scan, acx_ifnet_scan_period, acx_ifnet_callout_next_scan, (void *)ifp);
    }
}

/*
 * handle device ioctl
 *
 * notes:
 *  we trap three ioctls and pass the rest on to ieee802.11.
 */  
int
acx_ifnet_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
    struct acx_softc *sc = ifp->if_softc;
    int error = 0;
    struct ieee80211req *ieee80211req;
    //struct ieee80211com *ic= &sc->ic;

    switch (cmd)
    {
        case ACX_IOCTL_SGIOC:                           /* private ioctl */
            error = acx_ioctl(ifp, cmd, data);
            break;

        case SIOCSIFFLAGS:                              /* interface up/down */
            if( ifp->if_flags & IFF_UP )
                acx_ifnet_up(ifp);                      /* bring the interface up */
            else
                acx_ifnet_down(ifp);                    /* take the interface down */
            break;
        
        case SIOCADDMULTI:
        case SIOCDELMULTI:                              /* XXX unhandled multicast add/delete */
            break;

        default:
            error = ieee80211_ioctl(ifp, cmd, data);
            if( cmd==SIOCS80211 )                       /* pass wep update to the firmware */
            {
                ieee80211req= (struct ieee80211req *) data;
                if(ieee80211req->i_val>=0 && ieee80211req->i_val<IEEE80211_WEP_NKID) switch( ieee80211req->i_type )
                {
                    case IEEE80211_IOC_WEPKEY:          /* cache wep key and program device */
                        memcpy(&sc->wepkey[ieee80211req->i_val].key, ieee80211req->i_data, ieee80211req->i_len);
                        sc->wepkey[ieee80211req->i_val].len= ieee80211req->i_len;
                        acx_cmd_set_wep_key(sc, ieee80211req->i_val);
                        break;

                    case IEEE80211_IOC_WEPTXKEY:        /* cache tx key and program device */
                        acx_cmd_set_wep_tx_key(sc, sc->wepindex=ieee80211req->i_val );
                        break;
                }
            }
            if( error==ENETRESET) // XXX do not reset on fixed rate change, else association wedges
            {
                if( ifp->if_flags & IFF_RUNNING )
                    acx_ifnet_down(ifp);
                acx_ifnet_init(sc);
                error= 0;
            }
            break;
    }
    return error;
}

/*
 * setup available rates
 */
void
acx_ifnet_setrates(struct ieee80211com *ic)
{
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_nrates = 5;
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[0] = 2;   /* 1Mbps */
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[1] = 4;   /* 2Mbps */
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[2] = 11;  /* 5.5Mbps */
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[3] = 22;  /* 11Mbps */
    ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[4] = 44;  /* 22Mbps */
}

/*
 * setup available channels
 */
void
acx_ifnet_setchannels(struct ieee80211com *ic, u_int8_t low, u_int8_t high)
{
    u_int8_t i;

    for(i = low; i <= high; i++)
    {
        ic->ic_channels[i].ic_freq = ieee80211_ieee2mhz(i, IEEE80211_CHAN_B);
        ic->ic_channels[i].ic_flags = IEEE80211_CHAN_B;
    }
}

/*
 * bring the interface up and start scanning
 */
void
acx_ifnet_init(void *arg)
{
    struct acx_softc *sc;
    struct ieee80211com *ic;
    struct ifnet *ifp;

    sc = (struct acx_softc *)arg;
    ic = &sc->ic;
    ifp = &ic->ic_if;
    
    /* exit if interface is already running */
    if( ifp->if_flags & IFF_RUNNING )
        return;

    /*
     * enable command queue irqs
     *
     * notes:
     *  this has to be called as the interface comes
     *  online, and cannot be done before this point
     *  as we could be inside the boot sequence.
     *  it doesn't matter that we call this function
     *  more than once.
     */
    acx_dev_irq_on(sc);
        
    /* bring the interface up */
    acx_ifnet_up(ifp);
}

/*
 * bring the interface up and available for usage
 */
void
acx_ifnet_up(struct ifnet *ifp)
{
    struct acx_softc *sc = ifp->if_softc;
    struct ieee80211com *ic = &sc->ic;
    int rx_config=0;

    switch(ic->ic_opmode)
    {
        case IEEE80211_M_STA:
            rx_config= ACX_CMD_RXCFG_BSS;
            break;
        case IEEE80211_M_IBSS:            
            rx_config= ACX_CMD_RXCFG_IBSS;
            break;
        case IEEE80211_M_HOSTAP:
        case IEEE80211_M_MONITOR:
        case IEEE80211_M_AHDEMO:
            break;
    }

    /* turn on promiscuous mode */
//    if( ifp->if_flags & IFF_PROMISC )
//        rx_config &= ~ACX_CMD_RXCFG_UNPROMISCUOUS;

    /* set receiver config */
    acx_cmd_set_rx_config(sc, rx_config);
            
    acx_packet_txrate_reset(sc);        /* reset tx rate */

    /* mark the interface as running */
    ifp->if_flags |= IFF_RUNNING;
    ifp->if_flags &= ~IFF_OACTIVE;

    /* resume scanning if not running */
    if( ic->ic_state!=IEEE80211_S_RUN)
        acx_ifnet_start_scan(ifp);
}

/*
 * take the interface down and unavailble for usage
 */
void
acx_ifnet_down(struct ifnet *ifp)
{
    struct acx_softc *sc = ifp->if_softc;
    struct ieee80211com *ic = &sc->ic;

    /* mark the interface as not running */
    ifp->if_flags &= ~IFF_RUNNING;
    ifp->if_flags |= IFF_OACTIVE;

    /* stop interface */
    ic->ic_state = IEEE80211_S_INIT;
}

/*
 * station new state
 *
 * notes:
 *  this temporary hack is here to reassociate when the ieee80211
 *  state wedges at ASSOC->AUTH
 */
int
acx_ifnet_newstate(struct ieee80211com *ic, enum ieee80211_state newstate, int arg)
{
    struct ifnet *ifp= &ic->ic_if;
    struct acx_softc *sc= ifp->if_softc;
    struct ieee80211_node *ni;
    char *rates_string;

    if( newstate==IEEE80211_S_AUTH )
    {
        if(ic->ic_opmode == IEEE80211_M_STA)    /* joined infrastructure */
        {
            acx_packet_txrate_reset(sc);        /* reset tx rate */

            ni= ic->ic_bss; /* XXX should probably reference in case it disappears somehow */

            rates_string= acx_util_rates(&ni->ni_rates);
            device_printf(sc->dev, "AP rates:%s\n", rates_string);
            FREE(rates_string, M_TEMP);

            device_printf(sc->dev, "AP address: %s\n", ether_sprintf(ni->ni_macaddr) );
            
            if(ic->ic_state == IEEE80211_S_ASSOC)
                ic->ic_state = IEEE80211_S_RUN; /* rejoin */
        }
    }
    else if(newstate==IEEE80211_S_RUN)
    {
        if(ic->ic_opmode == IEEE80211_M_IBSS)   /* joined adhoc */
        {
        }
    }
    return(*sc->ic_newstate)(ic, newstate, arg);
}
