/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxpacket.c,v 1.235 2004/02/26 10:32:56 darron Exp $
 */

/* functions that perform I/O */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxreg.h"
#include "if_acxcmd.h"
#include "if_acxifnet.h"

/* functions */
#include "if_acxreg_proto.h"
#include "if_acxcmd_proto.h"
#include "if_acxioctl_proto.h"
#include "if_acxpacket_proto.h"
#include "if_acxutil_proto.h"

/* global sysctl vars */
int acx_packet_starts_fast=ACX_PACKET_STARTS_FAST;  /* initial tx rate is fast? */
int acx_packet_threshold=ACX_PACKET_THRESHOLD;      /* number of no error packets sample size */
int acx_packet_min_time=ACX_PACKET_MIN_TIME;        /* min sample time */
int acx_packet_max_time=ACX_PACKET_MAX_TIME;        /* max sample time */

/*
 * setup dma
 */
int
acx_packet_attach(struct acx_softc *sc)
{
    /* initialise locks */
    mtx_init(&sc->rx_lock, device_get_nameunit(sc->dev), "packet rx lock", MTX_DEF);
    mtx_init(&sc->tx_lock, device_get_nameunit(sc->dev), "packet tx lock", MTX_DEF);
    
    /*
     * create dma tag for tx/rx descriptors
     *
     * notes:
     *  tx and rx descriptors are of equal size.
     *  tx and rx rings are of equal size.
     *  
     *  there are two segments:
     *
     *  1) tx packet descriptors
     *  2) rx packet descriptors
     *  
     *  the tags are setup assuming that the parameters given are
     *  templates for each segment. it's not quite clear from the
     *  man page how and what we should be doing here.
     */
    if( bus_dma_tag_create(
        /* parent */        NULL,
        /* alignment */     PAGE_SIZE,
        /* boundary */      0,
        /* lowaddr */       BUS_SPACE_MAXADDR_32BIT,
        /* high addr */     BUS_SPACE_MAXADDR,
        /* filtfunc */      NULL,
        /* filtfuncarg */   NULL,
        /* maxsize */       ACX_PACKET_DESCSIZE * ACX_PACKET_RINGSIZE,
        /* nsegments */     1,
        /* maxsegsz */      ACX_PACKET_DESCSIZE * ACX_PACKET_RINGSIZE,
        /* flags */         0,
        /* lockfunc */      NULL,
        /* lockfuncarg */   NULL,
        /* dmat */          &sc->dma_tag) != 0)
    {
        device_printf(sc->dev, "failed to create dma tag\n");
        goto error;
    }

    /*
     * create dma tag for rx/tx buffers
     */
    if( bus_dma_tag_create(
        /* parent */        NULL,
        /* alignment */     PAGE_SIZE,
        /* boundary */      0,
        /* lowaddr */       BUS_SPACE_MAXADDR_32BIT,
        /* high addr */     BUS_SPACE_MAXADDR,
        /* filtfunc */      NULL,
        /* filtfuncarg */   NULL,
        /* maxsize */       MCLBYTES,
        /* nsegments */     1,
        /* maxsegsz */      MCLBYTES,
        /* flags */         0,
        /* lockfunc */      NULL,
        /* lockfuncarg */   NULL,
        /* dmat */          &sc->dma_mtag) != 0)
    {
        device_printf(sc->dev, "failed to create dma tag\n");
        goto error_dmatagdestroy;
    }

    /*
     * setup host tx packet descriptors
     *
     * notes:
     *  it would appear from observation that firmware descriptors
     *  are pointers to host based fragment descriptors which means
     *  that a host based ring is not how we should set things up.
     *  should this ever be investigated and proven true, then host
     *  base ring setup can be abandoned, and mbuf chains may be
     *  dma'd without using the current technique of concatenating
     *  a cluster into a fixed size outbound buffer. nb. in this
     *  scenario each mbuf item would need a tx host descriptor setup
     *  for it, so we would use more than one host descriptor per
     *  packet rather than using just the one as we do now.
     */
    if( bus_dmamap_create(
        /* dmat */          sc->dma_tag,
        /* flags */         0,
        /* mapp */          &sc->dma_map_txdesc) != 0 )
    {
        device_printf(sc->dev, "failed to initialise dma map\n");
        goto error_dmamtagdestroy;
    }
    if( bus_dmamem_alloc(
        /* dmat */          sc->dma_tag,
        /* vaddr */         (void **)&sc->dma_vaddr_txdesc,
        /* flags */         BUS_DMA_NOWAIT | BUS_DMA_ZERO,
        /* mapp */          &sc->dma_map_txdesc) != 0 )
    {
        device_printf(sc->dev, "failed to alloc dma memory\n");
        goto error_dmamapdestroy_txdesc;
    }
    if( bus_dmamap_load(
        /* dmat */          sc->dma_tag,
        /* map */           sc->dma_map_txdesc,
        /* buf */           sc->dma_vaddr_txdesc,
        /* buflen */        ACX_PACKET_DESCSIZE * ACX_PACKET_RINGSIZE,
        /* callback */      acx_packet_map_addr,
        /* callback_arg */  &sc->dma_paddr_txdesc,
        /* flags */         BUS_DMA_NOWAIT) != 0 )
    {
        device_printf(sc->dev, "failed to determine dma memory address\n");
        goto error_dmafree_txdesc;
    }
    
    /*
     * setup host rx packet descriptors
     *
     * notes:
     *  host based desciptors link directly into mbufs and dma should be
     *  optimised as best it can be.
     *  on a high latency machine (ie slow) the default ring size may be
     *  sub optimal and packets dropped. under that circumstance the ring
     *  size may be comfortably incresed to 32 or even 48 without any
     *  noticeable side effect.
     */
    if( bus_dmamap_create(
        /* dmat */          sc->dma_tag,
        /* flags */         0,
        /* mapp */          &sc->dma_map_rxdesc) != 0 )
    {
        device_printf(sc->dev, "failed to initialise dma map\n");
        goto error_dmaunload_txdesc;
    }
    if( bus_dmamem_alloc(
        /* dmat */          sc->dma_tag,
        /* vaddr */         (void **)&sc->dma_vaddr_rxdesc,
        /* flags */         BUS_DMA_NOWAIT | BUS_DMA_ZERO,
        /* mapp */          &sc->dma_map_rxdesc) != 0 )
    {
        device_printf(sc->dev, "failed to alloc dma memory\n");
        goto error_dmamapdestroy_rxdesc;
    }
    if( bus_dmamap_load(
        /* dmat */          sc->dma_tag,
        /* map */           sc->dma_map_rxdesc,
        /* buf */           sc->dma_vaddr_rxdesc,
        /* buflen */        ACX_PACKET_DESCSIZE * ACX_PACKET_RINGSIZE,
        /* callback */      acx_packet_map_addr,
        /* callback_arg */  &sc->dma_paddr_rxdesc,
        /* flags */         BUS_DMA_NOWAIT) != 0 )
    {
        device_printf(sc->dev, "failed to determine dma memory address\n");
        goto error_dmafree_rxdesc;
    }

    /*
     * initialise tx descriptor table and allocate bufs
     */
    if( acx_packet_init_txdesc(sc) != 0 )
        goto error_dmaunload_rxdesc;

    /*
     * initialise rx descriptor table and allocate mbufs
     */
    if( acx_packet_init_rxdesc(sc) == 0 )
        return 0;
    
    /*
     * exceptions
     */
    acx_packet_delete_bufmaps(sc, sc->txbuf);
error_dmaunload_rxdesc:
    bus_dmamap_unload(sc->dma_tag, sc->dma_map_rxdesc);
error_dmafree_rxdesc:
    bus_dmamem_free(sc->dma_tag, sc->dma_vaddr_rxdesc, sc->dma_map_rxdesc);
error_dmamapdestroy_rxdesc:
    bus_dmamap_destroy(sc->dma_tag, sc->dma_map_rxdesc);

error_dmaunload_txdesc:
    bus_dmamap_unload(sc->dma_tag, sc->dma_map_txdesc);
error_dmafree_txdesc:
    bus_dmamem_free(sc->dma_tag, sc->dma_vaddr_txdesc, sc->dma_map_txdesc);
error_dmamapdestroy_txdesc:
    bus_dmamap_destroy(sc->dma_tag, sc->dma_map_txdesc);

error_dmamtagdestroy:
    bus_dma_tag_destroy(sc->dma_mtag);

error_dmatagdestroy:
    bus_dma_tag_destroy(sc->dma_tag);

error:
    return -1;
}

/*
 * callback for memory mapping
 */
void
acx_packet_map_addr(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
    bus_addr_t *dma_paddr= (bus_addr_t *) arg;

    (*dma_paddr)= (bus_addr_t) segs->ds_addr;
}

/*
 * clean up for detachment
 */
void
acx_packet_detach(struct acx_softc *sc)
{
    acx_packet_delete_bufmaps(sc, sc->rxbuf);
    acx_packet_delete_bufmaps(sc, sc->txbuf);

    bus_dmamap_unload(sc->dma_tag,  sc->dma_map_rxdesc);
    bus_dmamem_free(sc->dma_tag,    sc->dma_vaddr_rxdesc, sc->dma_map_rxdesc);
    bus_dmamap_destroy(sc->dma_tag, sc->dma_map_rxdesc);

    bus_dmamap_unload(sc->dma_tag,  sc->dma_map_txdesc);
    bus_dmamem_free(sc->dma_tag,    sc->dma_vaddr_txdesc, sc->dma_map_txdesc);
    bus_dmamap_destroy(sc->dma_tag, sc->dma_map_txdesc);

    bus_dma_tag_destroy(sc->dma_mtag);
    bus_dma_tag_destroy(sc->dma_tag);

    mtx_destroy(&sc->tx_lock);
    mtx_destroy(&sc->rx_lock);
}

/*
 * configure firmware/chip descriptor tables
 */
int
acx_packet_init(struct acx_softc *sc)
{
    struct acx_cmd_memory_map_param *memory_map;
    struct acx_cmd_memory_config_options_param *memory_options;
    struct acx_cmd_queue_config_param *queue_config;
    u_int32_t queue_start;
    u_int16_t total_blocks;

    /* initialise rxdesc queue irq handler (fast) */
    TASK_INIT( &sc->packet_task_rx,
        /* priority */ 0,
        /* function */ (task_fn_t *)acx_packet_task_rx,
        /* context  */ sc);
    
    /* initialise rxdesc queue irq handler (fast) */
    TASK_INIT( &sc->packet_task_tx,
        /* priority */ 0,
        /* function */ (task_fn_t *)acx_packet_task_tx,
        /* context  */ sc);

    /*
     * modify memory map for packets
     */
    MALLOC(memory_map, struct acx_cmd_memory_map_param *,
        sizeof(struct acx_cmd_memory_map_param), M_TEMP, M_WAITOK | M_ZERO);

    acx_cmd_get_memory_map(sc, memory_map);
    queue_start = memory_map->QueueStart = 4 + memory_map->PacketTemplateEnd;
    acx_cmd_set_memory_map(sc, memory_map);

    FREE(memory_map, M_TEMP);

    /*
     * setup queue configuration
     */
    MALLOC(queue_config, struct acx_cmd_queue_config_param *,
        sizeof(struct acx_cmd_queue_config_param), M_TEMP, M_WAITOK | M_ZERO);

    /*
     * set queue sizes
     */
    sc->txsize= ACX_PACKET_RINGSIZE * sizeof(struct acx_packet_firmware_txdesc);
    sc->rxsize= ACX_PACKET_RINGSIZE * sizeof(struct acx_packet_firmware_rxdesc);

    queue_config->AreaSize= sc->txsize + sc->rxsize + 8;

    /*
     * set the start of the firmware tx descriptor queue
     */
    sc->txoffset = queue_config->TxQueueStart = queue_start;

    /*
     * set the start of the firmware rx descriptor queue
     */
    sc->rxoffset = queue_config->RxQueueStart = (sc->txoffset + sc->txsize);

    /*
     * set the end of the firmware rx descriptor queue
     */
    queue_config->QueueEnd = (sc->rxoffset + sc->rxsize);

    /*
     * set the start of the next (unknown) firmware queue
     */
    queue_config->QueueEnd2= queue_config->QueueEnd + 8;

    queue_config->vald= 1;  /* unknown variable */
    queue_config->vale= 1;  /* unknown variable */
    queue_config->valj= 0;  /* unknown variable */

    /*
     * configure queue settings
     */
    acx_cmd_set_queue_config(sc, queue_config);

#ifdef IOCTL
    acx_debug_dump_queue_config(sc, queue_config);
#endif

    FREE(queue_config, M_TEMP);

    /*
     * initialise firmware descriptors with control information
     */
    acx_packet_init_firmware_rxdesc(sc);
    acx_packet_init_firmware_txdesc(sc);

    /*
     * modify memory map for end
     */
    MALLOC(memory_map, struct acx_cmd_memory_map_param *,
        sizeof(struct acx_cmd_memory_map_param), M_TEMP, M_WAITOK | M_ZERO);

    acx_cmd_get_memory_map(sc, memory_map);
    memory_map->MemoryMap1= (memory_map->QueueEnd + 0x1f + 4) & 0xffffffe0; /* fix to 32 byte boundary */
    acx_cmd_set_memory_map(sc, memory_map);

#ifdef IOCTL
    acx_debug_dump_memory_map(sc, memory_map);
#endif

    /*
     * set the dma block size
     */
    acx_cmd_set_block_size(sc, ACX_PACKET_BLOCKSIZE);

    /*
     * set memory config options
     */
    MALLOC(memory_options, struct acx_cmd_memory_config_options_param *,
        sizeof(struct acx_cmd_memory_config_options_param), M_TEMP, M_WAITOK | M_ZERO);

    acx_cmd_get_memory_map(sc, memory_map);

    /*
     * XXX explain what is going on here
     */
    total_blocks                        = (memory_map->MemoryMap2 - memory_map->MemoryMap1) / ACX_PACKET_BLOCKSIZE;
    memory_options->TxBlockNum          = total_blocks / 2;
    memory_options->RxBlockNum          = total_blocks - memory_options->TxBlockNum;
    memory_options->DMA_config          = 0x00030000;
    memory_options->PoolStartAddress    = sc->dma_paddr_rxdesc;
    memory_options->val0xc              = (memory_map->MemoryMap1 + 0x1f) & 0xffffffe0;
    memory_options->val0x10             = (memory_map->MemoryMap1 + memory_options->TxBlockNum * ACX_PACKET_BLOCKSIZE + 0x01f) & 0xffffffe0;

    acx_cmd_set_memory_config_options(sc, memory_options);

#ifdef IOCTL
    acx_debug_dump_memory_options(sc, memory_options);
#endif

    FREE(memory_options, M_TEMP);
    FREE(memory_map, M_TEMP);

    /*
     * init memory to start the mac
     */
    acx_cmd_init_memory(sc);

    return 0;
}

/*
 * init rx descriptor ring
 *
 * notes:
 *  initialise the host rx desciptor ring.
 *  linkage may not be a requirement and should be tested.
 */
int
acx_packet_init_rxdesc(struct acx_softc *sc)
{
    int error;
    int i;
    bus_addr_t dma_paddr_rxdesc= sc->dma_paddr_rxdesc;
    struct acx_packet_rxdesc *rxdesc;
    struct acx_packet_bufmap *rxbuf;

    /*
     * initialise ring position
     */
    sc->rx_ring=0;

    /*
     * setup descriptors 0..RINGSIZE-1
     */
    for(i=0; i<ACX_PACKET_RINGSIZE; i++)
    {
        rxdesc = &sc->dma_vaddr_rxdesc[i];
        rxbuf  = &sc->rxbuf[i];

        /* allocate an mbuf and data storage area */
        error = acx_packet_new_mbuf(sc, &rxdesc->data_phy, rxbuf);
        if(error != 0)
        {
            acx_packet_delete_bufmaps(sc, rxbuf);
            return error;
        }

        /* store physical address of descriptor */
        rxdesc->desc_phy= (struct acx_packet_rxdesc *)dma_paddr_rxdesc;

        /* advance to next physical address of packet descriptor */
        dma_paddr_rxdesc+= sizeof( struct acx_packet_rxdesc );

        /* store physical address of next descriptor */
        rxdesc->desc_phy_next= (struct acx_packet_rxdesc *)dma_paddr_rxdesc;

        /* setup control block */
        rxdesc->length  = MCLBYTES;
        rxdesc->Ctl     = 0;
        rxdesc->val0x28 = 2;
    }

    /* complete the ring with the last descriptor */
    sc->dma_vaddr_rxdesc[ACX_PACKET_RINGSIZE-1].desc_phy_next=
        (struct acx_packet_rxdesc *)sc->dma_paddr_rxdesc;

    /* sync rxdesc vram to physical ram */
    bus_dmamap_sync(sc->dma_tag, sc->dma_map_rxdesc, BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE);

    return 0;
}

/*
 * init firmware rx descriptor ring
 */
void
acx_packet_init_firmware_rxdesc(struct acx_softc *sc)
{
    int i;
    bus_size_t offset= sc->rxoffset;
    struct acx_packet_firmware_rxdesc *rxdesc;
    
    MALLOC(rxdesc, struct acx_packet_firmware_rxdesc *,
        sizeof(struct acx_packet_firmware_rxdesc), M_TEMP, M_WAITOK | M_ZERO);

    /* setup control block common settings */
    rxdesc->val0x28= 0xc;

    for(i=0; i<=ACX_PACKET_RINGSIZE-2; i++)
    {
        /* setup control block */
        rxdesc->pNextDesc= offset + sizeof(struct acx_packet_firmware_rxdesc);

        /* store rxdesc in chipram */
        bus_space_write_region_1( sc->mem2.bustag, sc->mem2.bushandle,
            offset, (u_int8_t *) rxdesc, sizeof(struct acx_packet_firmware_rxdesc) );

        /* advance rxdesc offset */
        offset+= sizeof(struct acx_packet_firmware_rxdesc);
    }

    /* setup control block */
    rxdesc->pNextDesc= sc->rxoffset;

    /* store rxdesc in chipram */
    bus_space_write_region_1( sc->mem2.bustag, sc->mem2.bushandle,
        offset, (u_int8_t *) rxdesc, sizeof(struct acx_packet_firmware_rxdesc) );

    FREE(rxdesc, M_TEMP);
}

/*
 * new mbuf
 */
int
acx_packet_new_mbuf(struct acx_softc *sc, u_int8_t **paddr, struct acx_packet_bufmap *bufmap)
{
    int error;

    bufmap->type= ACX_PACKET_TYPE_MBUF;
    /*
     * allocate mbuf and extermal data storage area (2K) see kern/subr_mbuf.c
     * this also maps the mbuf data pointer to the external area
     *
     * notes:
     *  in theory we could allocate our own data buffer then map that into
     *  the m_ext area and in so doing be able to reserve >2k of storage.
     */
    bufmap->buf.m= m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
    if(bufmap->buf.m == NULL)
        return ENOBUFS;

    /* set size of data storage area */
    bufmap->buf.m->m_len = bufmap->buf.m->m_pkthdr.len = MCLBYTES;

    /* create a dma map */
    error = bus_dmamap_create(sc->dma_mtag, 0, &bufmap->map);
    if(error)
    {
        m_freem(bufmap->buf.m);
        bufmap->buf.m=NULL;
        return error;
    }

    /* map mbuf data storage and determine physical address */
    error = bus_dmamap_load_mbuf(
        sc->dma_mtag,
        bufmap->map,
        bufmap->buf.m,
        acx_packet_map_mbuf,
        paddr,
        0);

    if(error)
    {
        bus_dmamap_destroy(sc->dma_mtag, bufmap->map);
        m_freem(bufmap->buf.m);
        bufmap->buf.m=NULL;
        return error;
    }
    return 0;
}
void
acx_packet_map_mbuf(void *arg, bus_dma_segment_t *segs, int nseg,
    bus_size_t mapsize, int error)
{
    bus_addr_t *dma_paddr= (bus_addr_t *) arg;
    
    (*dma_paddr)= (bus_addr_t) segs->ds_addr;
}

/*
 * new buf
 */
int
acx_packet_new_buf(struct acx_softc *sc, u_int8_t **paddr, struct acx_packet_bufmap *bufmap)
{
    bufmap->type= ACX_PACKET_TYPE_BUF;

    if( bus_dmamap_create(
        sc->dma_mtag,
        0,
        &bufmap->map) != 0 )
    {
        return -1;
    }

    if( bus_dmamem_alloc(
        sc->dma_mtag,
        (void **)&bufmap->buf.b,
        BUS_DMA_NOWAIT | BUS_DMA_ZERO,
        &bufmap->map) != 0 )
    {
        bus_dmamap_destroy(sc->dma_mtag, bufmap->map);
        return -1;
    }

    if( bus_dmamap_load(
        /* dmat */          sc->dma_mtag,
        /* map */           bufmap->map,
        /* buf */           bufmap->buf.b,
        /* buflen */        MCLBYTES,
        /* callback */      acx_packet_map_addr,
        /* callback_arg */  paddr,
        /* flags */         BUS_DMA_NOWAIT) != 0 )
    {
        bus_dmamap_destroy(sc->dma_mtag, bufmap->map);
        bus_dmamem_free(sc->dma_mtag, bufmap->buf.b, bufmap->map);
        return -1;
    }
    return 0;
}

/*
 * delete buf/mbuf and map
 */
void
acx_packet_delete_bufmap(struct acx_softc *sc, struct acx_packet_bufmap *bufmap)
{
    /*
     * release map
     *
     * notes:
     *  on i386 maps are not usually created
     */
    if( bufmap->map != NULL )
    {
        bus_dmamap_unload(sc->dma_mtag, bufmap->map);
        bus_dmamap_destroy(sc->dma_mtag, bufmap->map);
        bufmap->map= NULL;
    }
    if( bufmap->type==ACX_PACKET_TYPE_MBUF && bufmap->buf.m!=NULL )
    {
        m_freem(bufmap->buf.m);
        bufmap->buf.m = NULL;
    }
    else if( bufmap->type==ACX_PACKET_TYPE_BUF && bufmap->buf.b!=NULL )
    {
        bus_dmamap_destroy(sc->dma_mtag, bufmap->map);
        bus_dmamem_free(sc->dma_mtag, bufmap->buf.b, bufmap->map);
        bus_dmamap_unload(sc->dma_tag,  bufmap->map);
    }
}

/*
 * delete bufs/mbufs and maps
 */
void
acx_packet_delete_bufmaps(struct acx_softc *sc, struct acx_packet_bufmap *bufmap)
{
    int i;

    for(i=0; i<ACX_PACKET_RINGSIZE; i++)
        acx_packet_delete_bufmap(sc, &bufmap[i]);
}

/*
 * init tx descriptor ring
 *
 * notes:
 *  initialise the host tx desciptor ring.
 *  linkage may not be a requirement and should be tested.
 *  in fact, none of this setup may actually be required.
 *  to test, fragment descriptors should be created on the
 *  fly or preallocated in a pool to chain entries from
 *  an mbuf cluster. the first entry in this chain would
 *  be liked from the firmware descriptor.
 */
int
acx_packet_init_txdesc(struct acx_softc *sc)
{
    int error;
    int i;
    bus_addr_t dma_paddr_txdesc= sc->dma_paddr_txdesc;
    struct acx_packet_txdesc *txdesc;
    struct acx_packet_bufmap *txbuf;

    /*
     * initialise ring position
     */
    sc->tx_ring=0;

    /*
     * setup descriptors 0..RINGSIZE-1
     */
    for(i=0; i<ACX_PACKET_RINGSIZE; i++)
    {
        txdesc = &sc->dma_vaddr_txdesc[i];
        txbuf  = &sc->txbuf[i];
        
        /* allocate a buf storage area */
        error = acx_packet_new_buf(sc, &txdesc->data_phy, txbuf);
        if(error != 0)
        {
            acx_packet_delete_bufmaps(sc, txbuf);
            return error;
        }

        /* store physical address of descriptor */
        txdesc->desc_phy  = (struct acx_packet_txdesc *)dma_paddr_txdesc;

        /* advance to next physical address of packet descriptor */
        dma_paddr_txdesc  += sizeof( struct acx_packet_txdesc );
        
        /* store physical address of next descriptor */
        txdesc->desc_phy_next  = (struct acx_packet_txdesc *)dma_paddr_txdesc;

        /* setup control block */
        txdesc->length = 0;
        txdesc->Ctl    = ACX_PACKET_CTL_FREE;
        txdesc->val0x4 = 0;
    }

    /* complete the ring with the last descriptor */
    sc->dma_vaddr_txdesc[ACX_PACKET_RINGSIZE-1].desc_phy_next=
        (struct acx_packet_txdesc *)sc->dma_paddr_txdesc;

    /* sync txdesc vram to physical ram */
    bus_dmamap_sync(sc->dma_tag, sc->dma_map_txdesc, BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE);

    return 0;
}

/*
 * init firmware tx descriptor ring
 */
void
acx_packet_init_firmware_txdesc(struct acx_softc *sc)
{
    int i;
    bus_size_t offset= sc->txoffset;
    struct acx_packet_firmware_txdesc *txdesc;
    bus_addr_t dma_paddr_txdesc= sc->dma_paddr_txdesc;

    MALLOC(txdesc, struct acx_packet_firmware_txdesc *,
        sizeof(struct acx_packet_firmware_txdesc), M_TEMP, M_WAITOK | M_ZERO);

    /* setup control block common settings */
    txdesc->Ctl= ACX_PACKET_CTL_INIT;

    for(i=0; i<=ACX_PACKET_RINGSIZE-2; i++)
    {
        /* setup control block */
        txdesc->pNextDesc= offset + sizeof(struct acx_packet_firmware_txdesc);
        txdesc->HostMemPtr= dma_paddr_txdesc;

        /* store txdesc in chipram */
        bus_space_write_region_1( sc->mem2.bustag, sc->mem2.bushandle,
            offset, (u_int8_t *) txdesc, sizeof(struct acx_packet_firmware_txdesc) );

        /* advance txdesc offset */
        offset+= sizeof(struct acx_packet_firmware_txdesc);

        /* advance physical tx descriptor offset */
        dma_paddr_txdesc+= sizeof(struct acx_packet_txdesc);
    }
    
    /* setup control block */
    txdesc->pNextDesc=  sc->txoffset;
    txdesc->HostMemPtr= dma_paddr_txdesc;

    /* store txdesc in chipram */
    bus_space_write_region_1( sc->mem2.bustag, sc->mem2.bushandle,
        offset, (u_int8_t *) txdesc, sizeof(struct acx_packet_firmware_txdesc) );

    FREE(txdesc, M_TEMP);
}
                
/*
 * RX ring irq task handler
 */
void
acx_packet_task_rx(struct acx_softc *sc, int count)
{
    struct ieee80211com *ic= &sc->ic;
    struct acx_packet_rxdesc *rxdesc;
    struct acx_packet_bufmap *rxbuf;
    struct acx_packet_tiheader *tiheader;
    struct ieee80211_frame *frame;
    struct ieee80211_node *ni;
    u_int8_t rxtest, rx_full=0, *payload;
    int len, payloadlen;

    mtx_lock(&sc->rx_lock);

    /* sync desriptor table */
    bus_dmamap_sync(sc->dma_tag, sc->dma_map_rxdesc, BUS_DMASYNC_POSTREAD|BUS_DMASYNC_POSTWRITE);
                
    /*
     * traverse the ring
     *
     * notes:
     *  we remember our last ring position, and when entering here skip up to
     *  the first packet in the queue, we process all packets from that point
     *  on until we either return to our original position, or the next packet
     *  is empty.
     */
    rxtest= sc->rx_ring;
    do
    {
        rxdesc = &sc->dma_vaddr_rxdesc[sc->rx_ring];

        /* test this descriptor for data */
        if( (rxdesc->Ctl     & ACX_PACKET_CTL_FREE) &&  /* descriptor now free, but full */
            (rxdesc->val0x14 & 0x80000000) )
        {
            /* update descriptor usage count */
            rx_full++;

            rxbuf = &sc->rxbuf[sc->rx_ring];

            if(rxbuf->buf.m==NULL)
            {
                device_printf(sc->dev, "acx_packet_task_rx: problem: rxbuf->mbuf==NULL\n");
            }
            else
            {
                /* sync mbuf data storage buffer to vram */
                bus_dmamap_sync(sc->dma_mtag, rxbuf->map, BUS_DMASYNC_POSTREAD);

                tiheader= mtod(rxbuf->buf.m, struct acx_packet_tiheader *);

                acx_ioctl_update_rx_stats(sc, tiheader);                /* update RX statistics */
            
                len= (tiheader->status & 0xfff);                        /* frame/payload length */

                if( len>=sizeof(struct ieee80211_frame_min) &&
                        len<MCLBYTES)                                   /* we have some data */
                {
                    ACX_IOCTL_UPDATE_RX_BYTES(sc, len);
                    
                    m_adj( rxbuf->buf.m, sizeof(struct acx_packet_tiheader) );
                    frame= mtod(rxbuf->buf.m, struct ieee80211_frame *);

                    if( frame->i_fc[1] & IEEE80211_FC1_WEP )
                    {
                        frame->i_fc[1]&= ~IEEE80211_FC1_WEP;            /* disable software WEP */

                        /*
                         * strip wep header and footer (4 bytes before payload and 4 bytes after)
                         *
                         * XXX
                         *  it would be better to shift the frame header down rather than move the payload up
                         */

                        /* ieee80211_frame_addr4 */
                        if ( (frame->i_fc[1] & IEEE80211_FC1_DIR_MASK) == IEEE80211_FC1_DIR_DSTODS )
                        {
                            payloadlen= len - sizeof(struct ieee80211_frame_addr4) - 8;
                            payload   = (u_int8_t *)frame + sizeof(struct ieee80211_frame_addr4);
                        }
                        else /* ieee80211_frame */
                        {
                            payloadlen= len - sizeof(struct ieee80211_frame) - 8;
                            payload   = (u_int8_t *)frame + sizeof(struct ieee80211_frame);
                        }
                        memcpy(payload, payload+4, payloadlen);
                        len-=8;
                    }
                    
                    rxbuf->buf.m->m_len = rxbuf->buf.m->m_pkthdr.len=len;
                    rxbuf->buf.m->m_pkthdr.rcvif = &ic->ic_if;

                    /* reference node */
                    if(ic->ic_opmode == IEEE80211_M_STA)
                        ni= ieee80211_ref_node(ic->ic_bss);
                    else
                    {
                        ni= ieee80211_find_node(ic, frame->i_addr2);
                        if(ni==NULL)
                            ni= ieee80211_ref_node(ic->ic_bss);
                    }

                    //acx_debug_packet(sc,"IN",tiheader,frame);

                    /* send packet to ieee 802.11 layer */
                    ieee80211_input(&ic->ic_if, rxbuf->buf.m, ni, tiheader->level, tiheader->time);

                    /* dereference node (deleted when reference count is zero and not us) */
                    if(ni==ic->ic_bss)
                        ieee80211_unref_node(&ni);
                    else
                        ieee80211_free_node(ic, ni);

                    /* ieee80211_input has freed the mbuf */
                    rxbuf->buf.m=NULL;
            
                    /* delete map */
                    acx_packet_delete_bufmap(sc, rxbuf);
                }
                else
                    ACX_IOCTL_UPDATE_RX_ERRORS(sc); /* erroneous packet counter */
            }

            /* allocate a new mbuf if needed */
            if(rxbuf->buf.m==NULL)
            {
                if(acx_packet_new_mbuf(sc, &rxdesc->data_phy, rxbuf) != 0)
                    device_printf(sc->dev, "acx_packet_task_rx: problem: failed to allocate mbuf\n");
            }

            /* mark this rx descriptor free for further use */
            if(rxbuf->buf.m!=NULL)
            {
                rxdesc->Ctl    &= ~ACX_PACKET_CTL_FREE;
                rxdesc->val0x14 = 0;
                bus_dmamap_sync(sc->dma_tag, sc->dma_map_rxdesc, BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE);
            }
        
            /* advance to next descriptor */
            sc->rx_ring= (sc->rx_ring + 1) % ACX_PACKET_RINGSIZE;
            if(!( (sc->dma_vaddr_rxdesc[sc->rx_ring].Ctl & ACX_PACKET_CTL_FREE) &&
                        (sc->dma_vaddr_rxdesc[sc->rx_ring].val0x14 & 0x80000000) ) )
                break; /* the next packet is empty, terminate */
        }
        else
        {
            /* skip past this empty descriptor */
            sc->rx_ring= (sc->rx_ring + 1) % ACX_PACKET_RINGSIZE;
        }
    }
    while( rxtest != sc->rx_ring );

    /*
     * ack irq
     */
    WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0008);

    mtx_unlock(&sc->rx_lock);
}

/*
 * return the next free tx descriptor offset else NULL on error
 *
 * notes:
 *  this should be called with the tx lock.
 *  reusing a descriptor that is free, and just been used which
 *  hasn't yet produced a tx irq seems to wedge the chip, make
 *  sure we don't use one.
 */
u_int16_t
acx_packet_txnext(struct acx_softc *sc)
{
    u_int16_t offset;
    u_int8_t control;

    /* determine offset of next firmware tx descriptor */
    offset= sc->txoffset + sc->tx_ring * sizeof(struct acx_packet_firmware_txdesc);

    control= READB(&sc->mem2, offset + ACX_PACKET_REGB_CTL);

    /* determine if descriptor is free */
    if( (control & ACX_PACKET_CTL_FREE) == 0 || (control & ACX_PACKET_CTL_USED_FOR_TX)  )
        return 0; /* not available */
    
    return offset;
}

/*
 * send packet(s)
 *
 * XXX
 * in the event that a "continue" occurs in this section, we have left
 * a hole in the tx descriptor list, this is probably a bug, but it doesn't
 * seem to happen (or does it)? debugging enabled to detect this feature.
 */
void
acx_packet_tx(struct ifnet *ifp)
{
    struct acx_softc *sc= ifp->if_softc;
    struct ieee80211com *ic= &sc->ic;
    struct ieee80211_frame *header;
    struct ieee80211_node *ni;
    struct mbuf *mbuf=NULL;
    struct acx_packet_bufmap *txbuf;
    struct acx_packet_txdesc *txdesc;
    u_int8_t rate=20, control;
    u_int16_t packet_len;
    bus_size_t offset;

    mtx_lock(&sc->tx_lock);

    while( (&ic->ic_mgtq)->ifq_head!=NULL || (&ifp->if_snd)->ifq_head!=NULL )
    {
        /* determine next tx descriptor */
        offset= acx_packet_txnext(sc);
        if(offset==0)
        {
            ifp->if_flags|= IFF_OACTIVE;
            mtx_unlock(&sc->tx_lock);
            return;
        }

        /* determine packet to send */
        if( (&ic->ic_mgtq)->ifq_head != NULL )          /* management packet */
        {
            IF_DEQUEUE(&ic->ic_mgtq, mbuf);
            if( mbuf == NULL )
            {
                device_printf(sc->dev, "acx_packet_tx: IF_DEQUEUE (mgmt) ret NULL mbuf\n");
                continue;
            }

            ni= (struct ieee80211_node *)mbuf->m_pkthdr.rcvif;
            mbuf->m_pkthdr.rcvif = NULL;

            rate=20;    /* XXX acx firmware rate of 2Mbps for management
                           XXX what if this isn't in the neg rate set? */
        }
        else if( (&ifp->if_snd)->ifq_head != NULL )     /* data packet */
        {
            if(ic->ic_state!=IEEE80211_S_RUN || !(ifp->if_flags & IFF_UP))
            {
                mtx_unlock(&sc->tx_lock);
                return;
            }

            IF_DEQUEUE(&ifp->if_snd, mbuf);
            if( mbuf == NULL )
            {
                device_printf(sc->dev, "acx_packet_tx: IF_DEQUEUE (data) ret NULL mbuf\n");
                continue;
            }

            /* feed sniffer */
            BPF_MTAP(ifp, mbuf);

            mbuf= ieee80211_encap(ifp, mbuf, &ni);
            if( mbuf == NULL )
            {
                device_printf(sc->dev, "acx_packet_tx: ieee80211_encap ret NULL mbuf\n");
                continue;
            }

            /*
             * enable hardware wep, if on
             */
            if(ic->ic_flags & IEEE80211_F_WEPON)
            {
                header= mtod(mbuf, struct ieee80211_frame *);
                header->i_fc[1]|= IEEE80211_FC1_WEP;
            }
        
            /*
             * set data rate
             *
             * XXX work out what to do about about the fixed rate setting later
             */

            //if(ic->ic_fixed_rate>=0)    /* user wants a fixed rate  */
            //{
            //rate= ACX_PACKET_IEEE2ACXRATE( ic->ic_bss->ni_rates.rs_rates[ ic->ic_fixed_rate ] );
            //}
            //else                        /* autoselected rate */
            //{
                rate= ACX_PACKET_IEEE2ACXRATE( ic->ic_bss->ni_rates.rs_rates[ ic->ic_bss->ni_txrate ] );
            //}
        }

        if( mbuf->m_pkthdr.len < sizeof(struct ieee80211_frame))
        {
            device_printf(sc->dev, "acx_packet_tx: problem: dropping short packet: m_pkthdr.len %d\n", mbuf->m_pkthdr.len);
            m_freem(mbuf);
            continue;
        }
        if( mbuf->m_pkthdr.len > MCLBYTES)
        {
            device_printf(sc->dev, "acx_packet_tx: problem: dropping long packet: m_pkthdr.len %d\n", mbuf->m_pkthdr.len);
            m_freem(mbuf);
            continue;
        }

        /* feed sniffer */
        if (ic->ic_rawbpf)
            bpf_mtap(ic->ic_rawbpf, mbuf);

        /* copy mbuf cluster into buffer */
        txbuf= &sc->txbuf[sc->tx_ring];
        m_copydata(mbuf, 0, mbuf->m_pkthdr.len, txbuf->buf.b);

        packet_len= mbuf->m_pkthdr.len; // + IEEE80211_CRC_LEN;
        m_freem(mbuf);
        bus_dmamap_sync(sc->dma_mtag, txbuf->map, BUS_DMASYNC_PREREAD);

        ACX_IOCTL_UPDATE_TX_BYTES(sc, packet_len);

        /* setup host tx descriptor */
        txdesc            = &sc->dma_vaddr_txdesc[sc->tx_ring];
        txdesc->Ctl      &= ~ACX_PACKET_CTL_FREE;
        txdesc->length    = packet_len;
        bus_dmamap_sync(sc->dma_tag, sc->dma_map_txdesc, BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE);

        /* setup firmware tx descriptor */
        control =  READB(&sc->mem2, offset + ACX_PACKET_REGB_CTL);
        control|=  ACX_PACKET_CTL_AUTODMA | ACX_PACKET_CTL_RECLAIM | ACX_PACKET_CTL_FIRST_MPDU;
        if ((ic->ic_flags & IEEE80211_F_SHPREAMBLE) && (ni->ni_capinfo & IEEE80211_CAPINFO_SHORT_PREAMBLE))
            control|= ACX_PACKET_CTL_SHORT_PREAMBLE;
        control&= ~ACX_PACKET_CTL_FREE;
        WRITEB(&sc->mem2, offset + ACX_PACKET_REGB_CTL,    control);
        WRITEB(&sc->mem2, offset + ACX_PACKET_REGB_RATE,   rate);
        WRITEL(&sc->mem2, offset + ACX_PACKET_REGL_LENGTH, packet_len);
        WRITEL(&sc->mem2, offset + ACX_PACKET_REGL_TXTIME, time_second);

        /* advance to next tx descriptor */
        sc->tx_ring= (sc->tx_ring + 1) % ACX_PACKET_RINGSIZE;
        
        /* tell chip */
        WRITEW(&sc->mem1, ACX100_REG_7C, 0x4);
        
        /* set timer */
        ifp->if_timer= 5;
    }

    mtx_unlock(&sc->tx_lock);
}

/*
 * initialise tx rate (infrastructure)
 */
void
acx_packet_txrate_reset(struct acx_softc *sc)
{
    struct ieee80211com *ic= &sc->ic;
    struct ieee80211_node *ni;

    /* filtered rate table */
    u_int8_t rs_nrates=0;
    u_int8_t rs_rates[IEEE80211_RATE_MAXSIZE];
    int x, y;

    /* init rate control */
    sc->txrate_time= acx_packet_min_time;   /* time to wait before rate up */
    sc->txrate_window= time_second;         /* moving timestamp of rate event */
    sc->txrate_last= time_second;           /* lag for updating above */
    sc->txrate_sample= 0;                   /* number of samples */

    /* hack: the neg rate table may include rates we cannot perform, so
    * we will strip those out of the node */

    /* work out what rates we support */
    ni= ic->ic_bss; /* XXX should probably reference in case it disappears somehow */
    for(x=0; x < ni->ni_rates.rs_nrates; x++)
        for(y=0; y < ic->ic_sup_rates[IEEE80211_MODE_11B].rs_nrates; y++)
            if( ic->ic_sup_rates[IEEE80211_MODE_11B].rs_rates[y]==(ni->ni_rates.rs_rates[x]&IEEE80211_RATE_VAL) )
                    rs_rates[rs_nrates++]= ni->ni_rates.rs_rates[x];

    /* update node with rates we can do */
    for(x=0;x<rs_nrates;x++)
        ni->ni_rates.rs_rates[x]= rs_rates[x];
    ni->ni_rates.rs_nrates= rs_nrates;
    
    /* set initial tx rate */
    if(acx_packet_starts_fast)
        ic->ic_bss->ni_txrate=rs_nrates-1;  /* start at highest rate */
    else
        ic->ic_bss->ni_txrate=0;            /* start at lowest rate */
}

/*
 * adjust tx rate to AP
 *
 * notes:
 *  XXX should be per peer (we adjust the peer rate, but what is there
 *  are many peers (adhoc,hostap) ... find out, work out)
 *
 *  XXX sample vars should move the SC namespace directly
 */
void
acx_packet_txrate(struct acx_softc *sc, int packetcount, int errorcount)
{
    struct ieee80211com *ic= &sc->ic;
    int timelag;
    char *r1,*r2; /* rate strings */

    /* 
     * update change time accounting for idle periods
     *
     * this means we test for X seconds of actual tranmission rather than X seconds
     * or either transmission or not.
     */
    sc->txrate_window+= ( (time_second - sc->txrate_last) & 0xfffe );

    /* timestamp this tx event */
    sc->txrate_last= time_second;

    /* handle error condition */
    if(errorcount)
    {
        /* determine amount of time between now and last rate change */
        timelag= time_second - sc->txrate_window;

        /* clear tx okay sample counter */
        sc->txrate_sample= 0;
                
        /* timestamp event */
        sc->txrate_window= time_second;

        if(timelag>1)
        {
            /*
             * work out time to wait before next rate up
             *
             * notes:
             *  if the timelag is small, make the next rate up test long
             *  if the timelag is long, make the next rate up test short
             *  if the time is neither of the above, subtract the timelag
             *  from the long test time
             */
            if( timelag <= acx_packet_min_time )
                sc->txrate_time= acx_packet_max_time;
            else if( timelag >= acx_packet_max_time )
                sc->txrate_time= acx_packet_min_time;
            else
            {
                sc->txrate_time= acx_packet_max_time - timelag;
                if( sc->txrate_time < acx_packet_min_time)
                    sc->txrate_time= acx_packet_min_time;
            }

            if( ic->ic_bss->ni_txrate>0 ) /* rate down possible */
            {
                /* take the rate down */
                r1= acx_util_rate(ic->ic_bss->ni_rates.rs_rates[  ic->ic_bss->ni_txrate]);
                r2= acx_util_rate(ic->ic_bss->ni_rates.rs_rates[--ic->ic_bss->ni_txrate]);
                device_printf(sc->dev, "rate down (%s -> %s)\n", r1, r2);
                FREE(r1, M_TEMP);
                FREE(r2, M_TEMP);
            }
        }
    }

    /* handle no error condition */
    else if( (1 + ic->ic_bss->ni_txrate) < ic->ic_bss->ni_rates.rs_nrates ) /* rate up possible */
    {
        /* update sample counter */
        sc->txrate_sample+= packetcount;
        
        /* determine amount of time between now and last rate change */
        timelag= time_second - sc->txrate_window;

        /* if it has been over min wait time since the last rate change and we have
         * sent at least XXX packets, try rate up now */
        if( timelag > sc->txrate_time && sc->txrate_sample >= acx_packet_threshold )
        {
            /* clear sample counter */
            sc->txrate_sample= 0;

            /* timestamp event */
            sc->txrate_window= time_second;

            /* determine next wait time (half of the this current wait time) */
            timelag= sc->txrate_time >> 1;
            if(timelag < acx_packet_min_time)
                sc->txrate_time= acx_packet_min_time;
            else if(timelag > acx_packet_max_time)
                sc->txrate_time= acx_packet_max_time;
            else
                sc->txrate_time= timelag;
                
            /* take the rate up */
            r1= acx_util_rate(ic->ic_bss->ni_rates.rs_rates[  ic->ic_bss->ni_txrate]);
            r2= acx_util_rate(ic->ic_bss->ni_rates.rs_rates[++ic->ic_bss->ni_txrate]);
            device_printf(sc->dev, "rate up (%s -> %s)\n", r1, r2);
            FREE(r1, M_TEMP);
            FREE(r2, M_TEMP);
        }
    }
}

/*
 * clean up the ring, looking at errors
 */
void
acx_packet_task_tx(struct acx_softc *sc, int count)
{
    struct ieee80211com *ic= &sc->ic;
    struct ifnet *ifp= &ic->ic_if;
    bus_size_t offset;
    u_int8_t ix, control, tx_full=0;
    int errorcount=0, packetcount=0;

    /* LOCK */
    mtx_lock(&sc->tx_lock);

    /* next descriptor for TX (we will test this last) */
    ix = sc->tx_ring;
    do
    {
        /* advance descriptor */
        ix= (ix + 1) % ACX_PACKET_RINGSIZE;

        /* determine firmware offset of tx descriptor */
        offset= sc->txoffset + (ix * sizeof(struct acx_packet_firmware_txdesc) );

        /* inspect descriptor status */
        control= READB(&sc->mem2, offset + ACX_PACKET_REGB_CTL);

        if( (control & ACX_PACKET_CTL_FREE) == 0 )      /* descriptor not free */
        {
            tx_full++;
        }
        else if( control & ACX_PACKET_CTL_USED_FOR_TX ) /* descriptor free and just used */
        {
            /*
             * detect tx error
             *
             * notes:
             *  there are at least two types of error, at the moment we don't
             *  differentiate between them.
             */
            if( READB(&sc->mem2, offset + ACX_PACKET_REGB_ERR) )
            {
                /* update total tx error count */
                ACX_IOCTL_UPDATE_TX_ERRORS(sc);
                errorcount++;
            }
        
            /* update tx packet counter */
            ACX_IOCTL_UPDATE_TX_STATS(sc);
            packetcount++;

            /* free descriptor for reuse */
            WRITEB( &sc->mem2, offset + ACX_PACKET_REGB_CTL, ACX_PACKET_CTL_FREE );

            /* mark interface as available for use */
            ifp->if_flags&= ~IFF_OACTIVE;
    
            /* reset timer */
            ifp->if_timer= 0;
        }
    }
    while( ix!=sc->tx_ring ); /* while not iterated over ring once */

    /* adjust tx rate, passing in number of errors this irq */
    acx_packet_txrate(sc, packetcount, errorcount);

    /* ack irq */
    WRITEW(&sc->mem1, ACX100_IRQ_ACK, 0x0002);

    /* UNLOCK */
    mtx_unlock(&sc->tx_lock);

    /* continue processing TX queue */
    acx_packet_tx(ifp);
}

/*
 * detect tx time out, make a point of it, and do nothing (yet)
 */
void
acx_packet_watchdog(struct ifnet *ifp)
{
    struct acx_softc *sc= ifp->if_softc;

    device_printf(sc->dev, "acx_packet_watchdog: problem: timeout\n");
    
    ieee80211_watchdog(ifp);
}
