/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxpci.c,v 1.30 2003/12/29 04:10:26 darron Exp $
 */

/* pci probe and attach */

/*
 * although this has been updated to detect the acx111 device, it
 * would appear to be radically different in nature to the older
 * acx100. a new driver model may need developing to incorporate
 * both devices in the same driver.
 */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxdev.h"
#include "if_acxpci.h"

/* functions */
#include "if_acx_proto.h"
#include "if_acxpci_proto.h"

/* device ids */
static struct acx_pci_id acx_pci_ids[]= {
    {
        ACX_DEV_100, 0x104c, 0x8400,
        "Texas Instruments (TI) 802.11b+ 22Mbps Wireless Adapter"
    },
    /*
    {
        ACX_DEV_111, 0x104c, 0x9066,
        "Texas Instruments (TI) 802.11g 54Mbps Wireless Adapter"
    },
    */

    {-1, -1, -1, "(null)"}
};

static device_method_t acx_pci_methods[]=
{
    DEVMETHOD(device_probe,     acx_pci_probe),
    DEVMETHOD(device_attach,    acx_pci_attach),
    DEVMETHOD(device_detach,    acx_pci_detach),
    DEVMETHOD(device_shutdown,  acx_pci_shutdown),
    DEVMETHOD(device_suspend,   acx_pci_suspend),
    DEVMETHOD(device_resume,    acx_pci_resume),
    { 0, 0 }
};

static driver_t acx_pci_driver=
{
    "acx", acx_pci_methods, sizeof(struct acx_softc)
};

static devclass_t acx_pci_devclass;

DRIVER_MODULE(if_acx, pci, acx_pci_driver, acx_pci_devclass, 0, 0);
DRIVER_MODULE(if_acx, cardbus, acx_pci_driver, acx_pci_devclass, 0, 0);

/*
 * probe
 */
int
acx_pci_probe(device_t dev)
{
    int i=0;

    do
    {
        if( acx_pci_ids[i].vendor==pci_get_vendor(dev) && acx_pci_ids[i].device==pci_get_device(dev) )
        {
            device_set_desc(dev, acx_pci_ids[i].description);
            return 0;
        }
    }
    while( acx_pci_ids[++i].type!=-1 );

    return ENXIO; /* not a known device to us */
}

/*
 * attach acx device
 *
 * notes:
 *  called after a successful probe acknowledgement
 */
static int
acx_pci_attach(device_t dev)
{
    struct acx_softc *sc= (struct acx_softc *)device_get_softc(dev);
    int i, rid;

    bzero(sc, sizeof(struct acx_softc));
    sc->dev= dev;
    sc->unit= device_get_unit(dev);
    
    /* determine device */
    i=0;
    while( acx_pci_ids[i].device != pci_get_device(sc->dev) )
       i++;
    sc->hardware_type= acx_pci_ids[i].type;

    /* enable bus mastering */
    pci_enable_busmaster(sc->dev); 
    if(! (pci_read_config(sc->dev, PCIR_COMMAND, 4) & PCIM_CMD_BUSMASTEREN) )
    {
        device_printf(sc->dev, "acx_pci_attach: failed to enable bus mastering\n");
        goto error;
    }
    
    /* allocate device i/o */
    if( sc->hardware_type == ACX_DEV_100 )
    {
        if( acx_pci_attach_acx100(sc) != 0 ) goto error;
    }
    else if( sc->hardware_type == ACX_DEV_111 )
    {
        if( acx_pci_attach_acx111(sc) != 0 ) goto error;
    }

    /* allocate irq */
    rid= 0;
    if( (sc->irq= bus_alloc_resource(sc->dev, SYS_RES_IRQ, &rid, 0ul, ~0ul, 1, RF_SHAREABLE|RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach: failed to map interrupt\n");
        goto error_releasedev;
    }
    if(bus_setup_intr(sc->dev, sc->irq, INTR_TYPE_NET, acx_interrupt, sc, &sc->irqhandler) != 0 )
    {
        device_printf(sc->dev, "acx_pci_attach: failed to start interrupts\n");
        goto error_releaseirqres;
    }

    /*
     * attach device
     */
    if( acx_attach(sc) == 0 )
        return 0;

    /*
     * exceptions
     */
    bus_teardown_intr(sc->dev, sc->irq, sc->irqhandler);
error_releaseirqres:
    bus_release_resource(sc->dev, SYS_RES_IRQ, 0, sc->irq);

error_releasedev:
    if( sc->hardware_type == ACX_DEV_100)
        acx_pci_detach_acx100(sc);
    else if( sc->hardware_type == ACX_DEV_111)
        acx_pci_detach_acx111(sc);

error:
    return ENXIO;
}

static int
acx_pci_attach_acx100(struct acx_softc *sc)
{
    int rid;

    /* allocate i/o resource */
    pci_enable_io(sc->dev, SYS_RES_IOPORT);

    if(! (pci_read_config(sc->dev, PCIR_COMMAND, 4) & PCIM_CMD_PORTEN) )
    {
        device_printf(sc->dev, "acx_pci_attach_acx100: failed to enable i/o\n");
        return ENXIO;
    }

    rid= 0x10;
    if( (sc->io= bus_alloc_resource(sc->dev, SYS_RES_IOPORT, &rid, 0ul, ~0ul, 1, RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach_acx100: failed to map i/o\n");
        return ENXIO;
    }

    /* allocate memory resource 1 */
    pci_enable_io(sc->dev, SYS_RES_MEMORY);

    if(! (pci_read_config(sc->dev, PCIR_COMMAND, 4) & PCIM_CMD_MEMEN) )
    {
        device_printf(sc->dev, "acx_pci_attach_acx100: failed to enable memory\n");
        bus_release_resource(sc->dev, SYS_RES_IOPORT, 0x10, sc->io);
        return ENXIO;
    }

    rid= 0x14;
    if( (sc->mem1.res= bus_alloc_resource(sc->dev, SYS_RES_MEMORY, &rid, 0ul, ~0ul, 1, RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach_acx100: failed to map memory region 1\n");
        bus_release_resource(sc->dev, SYS_RES_IOPORT, 0x10, sc->io);
        return ENXIO;
    }
    sc->mem1.bustag=    rman_get_bustag(sc->mem1.res);
    sc->mem1.bushandle= rman_get_bushandle(sc->mem1.res);

    /* allocate memory resource 2 */
    rid= 0x18;
    if( (sc->mem2.res= bus_alloc_resource(sc->dev, SYS_RES_MEMORY, &rid, 0ul, ~0ul, 1, RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach_acx100: failed to map memory region 2\n");
        bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x14, sc->mem1.res);
        bus_release_resource(sc->dev, SYS_RES_IOPORT, 0x10, sc->io);
        return ENXIO;
    }
    sc->mem2.bustag=    rman_get_bustag(sc->mem2.res);
    sc->mem2.bushandle= rman_get_bushandle(sc->mem2.res);

    return 0;
}

static int
acx_pci_attach_acx111(struct acx_softc *sc)
{
    int rid;

    /* allocate memory resource 1 */
    pci_enable_io(sc->dev, SYS_RES_MEMORY);

    if(! (pci_read_config(sc->dev, PCIR_COMMAND, 4) & PCIM_CMD_MEMEN) )
    {
        device_printf(sc->dev, "acx_pci_attach_acx111: failed to enable memory\n");
        return ENXIO;
    }

    rid= 0x10;
    if( (sc->mem1.res= bus_alloc_resource(sc->dev, SYS_RES_MEMORY, &rid, 0ul, ~0ul, 1, RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach_acx111: failed to map memory region 1\n");
        return ENXIO;
    }
    sc->mem1.bustag=    rman_get_bustag(sc->mem1.res);
    sc->mem1.bushandle= rman_get_bushandle(sc->mem1.res);

    /* allocate memory resource 2 */
    rid= 0x14;
    if( (sc->mem2.res= bus_alloc_resource(sc->dev, SYS_RES_MEMORY, &rid, 0ul, ~0ul, 1, RF_ACTIVE)) == NULL )
    {
        device_printf(sc->dev, "acx_pci_attach_acx111: failed to map memory region 2\n");
        bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x10, sc->mem1.res);
        return ENXIO;
    }
    sc->mem2.bustag=    rman_get_bustag(sc->mem2.res);
    sc->mem2.bushandle= rman_get_bushandle(sc->mem2.res);

    return 0;
}

/*
 * detach
 */
static int
acx_pci_detach(device_t dev)
{
    struct acx_softc *sc= (struct acx_softc *)device_get_softc(dev);

    acx_detach(sc);

    /* stop and release irq resources */
    bus_teardown_intr(dev, sc->irq, sc->irqhandler);
    bus_release_resource(dev, SYS_RES_IRQ, 0, sc->irq);

    /* release i/o */
    if( sc->hardware_type==ACX_DEV_100)
        acx_pci_detach_acx100(sc);
    else if( sc->hardware_type==ACX_DEV_111)
        acx_pci_detach_acx111(sc);

    return 0;
}

static int
acx_pci_detach_acx100(struct acx_softc *sc)
{
    if(sc->mem2.res) bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x18, sc->mem2.res);
    if(sc->mem1.res) bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x14, sc->mem1.res);
    if(sc->io)       bus_release_resource(sc->dev, SYS_RES_IOPORT, 0x10, sc->io);

    return 0;
}

static int
acx_pci_detach_acx111(struct acx_softc *sc)
{
    if(sc->mem2.res) bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x14, sc->mem2.res);
    if(sc->mem1.res) bus_release_resource(sc->dev, SYS_RES_MEMORY, 0x10, sc->mem1.res);

    return 0;
}

/*
 * shutdown
 */
static int
acx_pci_shutdown(device_t dev)
{
    device_printf(dev, "acx_pci_shutdown: unhandled\n");
    return 0;
}

/*
 * suspend
 */
static int
acx_pci_suspend(device_t dev)
{
    device_printf(dev, "acx_pci_suspend: unhandled\n");
    return 0;
}

/*
 * resume
 */
static int
acx_pci_resume(device_t dev)
{
    device_printf(dev, "acx_pci_resume: unhandled\n");
    return 0;
}
