/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acxtemplate.c,v 1.23 2003/12/28 20:30:07 darron Exp $
 */

/*
 * functions to create and setup templates.
 *
 * these templates should setup beacons, probe responses and
 * tim for use in hostap/adhoc modes.
 */

/* system */
#include "if_acxsys.h"

/* declarations */
#include "if_acx.h"
#include "if_acxreg.h"
#include "if_acxcmd.h"

/* functions */
#include "if_acxtemplate_proto.h"
#include "if_acxreg_proto.h"
#include "if_acxcmd_proto.h"

/*
 * initialise templates
 *
 * notes:
 *  order may be important.
 *  memory map must be initialised prior to intialising templates
 *  else firmware corruption will occur (overwrite).
 */
void
acx_template_init(struct acx_softc *sc)
{
    struct acx_cmd_memory_map_param *memory_map;
    u_int8_t *packet;

    /*
     * modify memory map for templates
     */
    MALLOC(memory_map, struct acx_cmd_memory_map_param *,
        sizeof(struct acx_cmd_memory_map_param), M_TEMP, M_WAITOK | M_ZERO);

    acx_cmd_get_memory_map(sc, memory_map);
    memory_map->PacketTemplateStart= memory_map->WEPCacheEnd;
    memory_map->PacketTemplateEnd  = memory_map->WEPCacheEnd;
    acx_cmd_set_memory_map(sc, memory_map);

    FREE(memory_map, M_TEMP);

    MALLOC(packet, u_int8_t *, 0x100, M_TEMP, M_WAITOK | M_ZERO);

    /*
     * init beacon response template
     */
    acx_cmd_config_beacon(sc, packet, 0x54);
    
    /*
     * init tim (traffic indication map) template
     */
    acx_cmd_config_tim(sc, packet, 0x100);

    /*
     * init probe response template
     */
    acx_cmd_config_probe_response(sc, packet, 0x54);
    
    FREE(packet, M_TEMP);

    /*
     * populate templates
     */
    /*
    acx_template_set_tim(sc);
    acx_template_set_beacon(sc);
    acx_template_set_probe_response(sc);
    */
}

void
acx_template_set_tim(struct acx_softc *sc)
{
    u_int8_t *packet;
    
    MALLOC(packet, u_int8_t *, 0x100, M_TEMP, M_WAITOK | M_ZERO);
    
    packet[0]= 0x5;
    packet[1]= 0x4;

    acx_cmd_config_tim(sc, packet, 0x100);
    
    FREE(packet, M_TEMP);
}

void
acx_template_set_beacon(struct acx_softc *sc)
{
    int len;
    u_int8_t *packet;
    
    MALLOC(packet, u_int8_t *, 0x100, M_TEMP, M_WAITOK | M_ZERO);

    len= acx_template_get_beacon(sc, packet, IEEE80211_FC0_SUBTYPE_BEACON);
        
    acx_cmd_config_beacon(sc, packet, len);
    
    FREE(packet, M_TEMP);
}   

void
acx_template_set_probe_response(struct acx_softc *sc)
{
    int len;
    u_int8_t *packet;
    
    MALLOC(packet, u_int8_t *, 0x100, M_TEMP, M_WAITOK | M_ZERO);

    len= acx_template_get_beacon(sc, packet, IEEE80211_FC0_SUBTYPE_PROBE_RESP);

    acx_cmd_config_probe_response(sc, packet, len);
    
    FREE(packet, M_TEMP);
}   

int
acx_template_get_beacon(struct acx_softc *sc, u_int8_t *beacon, u_int8_t subtype)
{
    static u_int8_t broadcast[]= {0xff,0xff,0xff,0xff,0xff,0xff};
    static u_int8_t bssid[]=     {0x00,0x00,0x00,0x00,0x00,0x00};

    /* these may need setting differently */
    static u_int8_t ssid[]="ANY";
    static u_int8_t channel=0;

    int ix;

    /*
     * BEACON management packets
     *
     *      octet timestamp[8]
     *      octet beacon interval[2]
     *      octet capability information[2]
     *      information element
     *              octet elemid
     *              octet length
     *              octet information[length]
     */

    beacon[0]=subtype;                                          /* i_fc[0] */
    beacon[1]=0;                                                /* i_fc[1] */
    beacon[2]=0;                                                /* i_dur[0] */
    beacon[3]=0;                                                /* i_dur[1] */
    
    IEEE80211_ADDR_COPY(&beacon[4],  broadcast);                /* i_addr1 */
    IEEE80211_ADDR_COPY(&beacon[10], sc->hardware_macaddr);     /* i_addr2 */
    IEEE80211_ADDR_COPY(&beacon[16], bssid);                    /* i_addr3 */

    beacon[22]=0;                                               /* i_seq[0] */
    beacon[23]=0;                                               /* i_seq[1] */

    bzero(&beacon[24], 8);                                      /* timestamp */
    
    beacon[32]= 100;                                            /* beacon interval */
    beacon[33]= 0;

    beacon[34]= IEEE80211_CAPINFO_ESS |                         /* capabilities */
                IEEE80211_CAPINFO_IBSS|
                IEEE80211_CAPINFO_PRIVACY|
                IEEE80211_CAPINFO_SHORT_PREAMBLE;
    beacon[35]=0;

    beacon[36]= IEEE80211_ELEMID_SSID;                          /* ssid */
    beacon[37]= strlen(ssid); ix= 38;
    memcpy( &beacon[ix], ssid, strlen(ssid) ); ix+= strlen(ssid);

    beacon[ix++]= IEEE80211_ELEMID_RATES;                       /* supported rates */
    beacon[ix++]= 5;
    beacon[ix++]= 0x82; /* 1 Mbps */
    beacon[ix++]= 0x84; /* 2 Mbps */
    beacon[ix++]= 0x8b; /* 5.5 Mbps */
    beacon[ix++]= 0x96; /* 11 Mbps */
    beacon[ix++]= 0xac; /* 22 Mbps */

    beacon[ix++]= IEEE80211_ELEMID_DSPARMS;                     /* current channel */
    beacon[ix++]= 1;
    beacon[ix++]= channel;

    beacon[ix++]= 0;
    beacon[ix++]= 0;
    beacon[ix++]= 0;

    return ix;
}
