/*
 * Copyright (c) 2003 wlan.kewl.org Project
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: if_acx.h,v 1.98 2004/02/26 09:31:28 darron Exp $
 */

#ifndef _IF_ACX_H
#define _IF_ACX_H

#include "if_acxpacket.h"
#include "if_acxwep.h"
#include "if_acxdev.h"
#include "if_acxioctl.h"

/*
 * device instance variables
 */

/* memory resource */
struct acx_resource
{
    struct resource *res;
    bus_space_tag_t bustag;
    bus_space_handle_t bushandle;
};

/* driver resource */
struct acx_softc
{
    /*
     * hardware info
     */
    int16_t hardware_type;                          /* acx100, acx111 (unsupported) */
    u_int16_t hardware_version;                     /* radio version */
    u_int8_t hardware_macaddr[IEEE80211_ADDR_LEN];  /* mac addrress (high->low) */

    /* cached hardware defaults for certain parameters loaded after firmware init */
    u_int8_t ed_threshold;
    u_int8_t current_antenna;
    u_int8_t cca_mode;
    u_int8_t unused0;

    /*
     * firmware info
     *
     * XXX rx and tx sizes are fixed to be equal elsewhere, we should make them
     * either always equal or always independant.
     */
    char    firmware_version[0x18]; /* station version */  
    u_int16_t cmdoffset;            /* command mailbox */
    u_int16_t infoffset;            /* info mailbox */
    u_int32_t txoffset;             /* tx descriptor queue */
    u_int16_t txsize;               /* tx descriptor size */
    u_int32_t rxoffset;             /* rx descriptor queue */
    u_int16_t rxsize;               /* rx descriptor size */

    struct acx_ioctl_stats stats;   /* real time statistics */

    /*
     * dma tags
     */
    bus_dma_tag_t dma_tag;          /* descriptor dma tag */       
    bus_dma_tag_t dma_mtag;         /* mbuf dma tag */

    /*
     * dma maps, memory and pointers
     */

    /* tx */
    bus_dmamap_t dma_map_txdesc;
    bus_addr_t dma_paddr_txdesc;
    struct acx_packet_txdesc *dma_vaddr_txdesc;
    struct acx_packet_bufmap txbuf[ACX_PACKET_RINGSIZE];

    /* rx */
    bus_dmamap_t dma_map_rxdesc;
    bus_addr_t dma_paddr_rxdesc;
    struct acx_packet_rxdesc *dma_vaddr_rxdesc;
    struct acx_packet_bufmap rxbuf[ACX_PACKET_RINGSIZE];
    
    /*
     * packet control and task handler
     */
    u_int8_t                       rx_ring;     /* rx ring index */
    u_int8_t                       tx_ring;     /* tx ring index */
    struct task                    packet_task_rx;
    struct task                    packet_task_tx;

    int txrate_time;    /* time to wait for txrate evaluation */
    int txrate_window;  /* moving timestamp of last txrate evaluation (accounts for idle periods) */
    int txrate_last;    /* last tx timestamp to update above */
    int txrate_sample;  /* number of packets sent over the interface (a minimum is a requirement for rate up) */

    /* XXX
     * there is some form of locking pool which may be appropriate here
     */
    struct mtx                     rx_lock;
    struct mtx                     tx_lock;

    /*
     * eeprom info
     */
    u_int16_t eeprom_version;       /* eeprom bootstrap version */
    u_int16_t eeprom_domain;        /* eeprom regulatory domain */
    const char *eeprom_domainstr;   /* eeprom regulatory domain string */
    u_int8_t eeprom_domainlow;      /* eeprom regulatory channel low */
    u_int8_t eeprom_domainhigh;     /* eeprom regulatory channel high */

    /*
     * device info
     */
    device_t        dev;            /* device handle */
    u_int8_t        unit;           /* device unit id */

    /*
     * resources
     */
    struct resource *io;            /* i/o resource size 32 bytes */
    struct acx_resource mem1;       /* memory resource 1 size 4 Kbytes */
    struct acx_resource mem2;       /* memory resource 2 size 64 Kbytes */
    struct resource *irq;           /* irq resource */
    void            *irqhandler;    /* intererupt handler */
    int irq_enabled;            /* when set, irqs are active
                                       and non busy wait firmware
                                       commands are in effect */

    /*
     * lock
     *
     * XXX will this ever be used? come on, lock the irq, damnit.
     */
    struct mtx      lock;          /* thread lock unused at present */

    /*
     * command queue list and task handler
     */
    STAILQ_HEAD(noname,acx_cmd_entry)     cmd_head;
    struct task                     cmd_task_irq;
    struct task                     cmd_task_exec;
    struct mtx                      cmd_lock;

    /*
     * null if no command, else cmd_entry of running command
     *
     * notes:
     *  this entry is volatile, do not access it
     */
    struct acx_cmd_entry            *cmd;

    /*
     * wep keys
     *
     * XXX these are not compatible in size with net80211 at this time
     * which does not support 232/256 bit wep. we may write a userland
     * tool to manipulate these keys directly, otherwise they will
     * cache the net80211 entries when invoked by an ioctl.
     */
    struct acx_wep_key wepkey[IEEE80211_WEP_NKID];
    u_int16_t wepindex;

    /*
     * ethernet/802.11 interface
     */
    struct ieee80211com             ic;
    int (*ic_newstate)(struct ieee80211com *, enum ieee80211_state, int);
    struct callout                  ifnet_callout_scan;
    struct mtx                      ifnet_lock;/*UNUSED*/
};

#endif /* !_IF_ACX_H */
